# -*- coding: utf-8 -*-
"""Revolution Pi PLC program configuration."""
__author__ = "Sven Sager"
__copyright__ = "Copyright (C) 2023 Sven Sager"
__license__ = "GPLv2"

import gzip
import os
import tarfile
import zipfile
from shutil import rmtree
from tempfile import mkdtemp
from xmlrpc.client import Binary

from PyQt5 import QtCore, QtGui, QtWidgets

from . import helper
from . import proginit as pi
from .ui.revpiprogram_ui import Ui_diag_program


class RevPiProgram(QtWidgets.QDialog, Ui_diag_program):
    """Program options of RevPiPyLoad."""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setupUi(self)
        self.setFixedSize(self.size())

        self.dc = {}
        self.lst_files = []

        self.cbx_pictory.setEnabled(False)
        self.recover_gui()

    def _apply_acl(self):
        """Set availability of controls depending on ACL level."""

        # Setting properties require level 4
        self.cbb_plcprogram.setEnabled(helper.cm.xml_mode >= 4)
        self.txt_plcarguments.setEnabled(helper.cm.xml_mode >= 4)
        self.rbn_pythonversion_2.setEnabled(helper.cm.xml_mode >= 4)
        self.rbn_pythonversion_3.setEnabled(helper.cm.xml_mode >= 4)
        self.cbx_plcworkdir_set_uid.setEnabled(helper.cm.xml_mode >= 4)

        # Downloads require level 2
        self.btn_program_download.setEnabled(helper.cm.xml_mode >= 2)
        self.btn_pictory_download.setEnabled(helper.cm.xml_mode >= 2)
        self.btn_procimg_download.setEnabled(helper.cm.xml_mode >= 2)

        # Uploads require level 3
        self.btn_program_upload.setEnabled(helper.cm.xml_mode >= 3)
        self.btn_pictory_upload.setEnabled(helper.cm.xml_mode >= 3)

    def _changesdone(self):
        """
        Check for unsaved changes in dialog.

        :return: True, if unsaved changes was found
        """
        return self.cbb_plcprogram.currentText() != self.dc.get("plcprogram", "") or \
            self.txt_plcarguments.text() != self.dc.get("plcarguments", "") or \
            self.rbn_pythonversion_2.isChecked() != (self.dc.get("pythonversion", 3) == 2) or \
            self.rbn_pythonversion_3.isChecked() != (self.dc.get("pythonversion", 3) == 3) or \
            int(self.cbx_plcworkdir_set_uid.isChecked()) != self.dc.get("plcworkdir_set_uid", 0) or \
            self.sbx_plcprogram_watchdog.value() != self.dc.get("plcprogram_watchdog", 0)

    def _load_settings(self, files_only=False):
        """Load values to GUI widgets."""
        pi.logger.debug("RevPiProgram._load_settings")

        if files_only:
            mrk_program = self.cbb_plcprogram.currentText()
        else:
            mrk_program = ""

        self.cbb_plcprogram.clear()
        self.cbb_plcprogram.addItem("")

        self.lst_files.sort()
        is_in_list = False
        for file in self.lst_files:
            if file == ".placeholder":
                continue
            self.cbb_plcprogram.addItem(file)
            check = mrk_program or self.dc.get("plcprogram", "")
            if file == check:
                is_in_list = True

        if not is_in_list:
            pi.logger.warning("File {0} is not in list".format(mrk_program or self.dc.get("plcprogram", "")))

        if files_only:
            self.cbb_plcprogram.setCurrentText(mrk_program)
        else:
            self.cbb_plcprogram.setCurrentText(self.dc.get("plcprogram", ""))
            self.txt_plcarguments.setText(self.dc.get("plcarguments", ""))
            self.rbn_pythonversion_2.setChecked(self.dc.get("pythonversion", 3) == 2)
            self.rbn_pythonversion_3.setChecked(self.dc.get("pythonversion", 3) == 3)
            self.cbx_plcworkdir_set_uid.setChecked(bool(self.dc.get("plcworkdir_set_uid", 0)))
            self.sbx_plcprogram_watchdog.setValue(self.dc.get("plcprogram_watchdog", 0))

    def accept(self) -> None:
        # todo: After upload ask for restart pcl program?
        if not self._changesdone():
            super().accept()
            return

        if self.cbb_plcprogram.currentText() == "":
            QtWidgets.QMessageBox.critical(
                self, self.tr("Error"), self.tr(
                    "You have to select a start program, before uploading the "
                    "settings."
                )
            )
            return

        ask = QtWidgets.QMessageBox.question(
            self, self.tr("Question"), self.tr(
                "The settings will be set on the Revolution Pi now.\n\n"
                "If you made changes on the 'PCL Program' section, your plc "
                "program will restart now!"
            )
        ) == QtWidgets.QMessageBox.Yes

        if not ask:
            return

        self.dc["plcprogram"] = self.cbb_plcprogram.currentText()
        self.dc["plcarguments"] = self.txt_plcarguments.text()
        self.dc["pythonversion"] = 2 if self.rbn_pythonversion_2.isChecked() else 3
        self.dc["plcworkdir_set_uid"] = int(self.cbx_plcworkdir_set_uid.isChecked())
        self.dc["plcprogram_watchdog"] = self.sbx_plcprogram_watchdog.value()

        saved = helper.cm.call_remote_function(
            "set_config", self.dc, ask,
            default_value=False
        )

        if saved:
            super().accept()
        else:
            QtWidgets.QMessageBox.critical(
                self, self.tr("Error"), self.tr(
                    "The settings could not be saved on the Revolution Pi!\n"
                    "Try to save the values one mor time and check the log "
                    "files of RevPiPyLoad if the error rises again."
                )
            )

    def closeEvent(self, a0: QtGui.QCloseEvent) -> None:
        if self._changesdone():
            ask = QtWidgets.QMessageBox.question(
                self, self.tr("Question"), self.tr(
                    "Do you really want to quit? \nUnsaved changes will be lost."
                )
            ) == QtWidgets.QMessageBox.Yes

            if ask:
                self.reject()
            else:
                a0.ignore()

    def exec(self) -> int:
        # Reset class values
        if not helper.cm.connected:
            return QtWidgets.QDialog.Rejected

        self.dc = helper.cm.call_remote_function("get_config", default_value={})
        self.lst_files = helper.cm.call_remote_function("get_filelist", default_value=None)
        if len(self.dc) == 0 or self.lst_files is None:
            return QtWidgets.QDialog.Rejected

        self._load_settings()
        self._apply_acl()

        return super().exec()

    def reject(self) -> None:
        """Reject all sub windows and reload settings."""
        self._load_settings()
        super().reject()

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    # region #      REGION: PLC program

    def _upload_pictory(self, filename: str):
        """Upload piCtory configuration."""
        fh = open(filename, "rb")
        file_binary = Binary(fh.read())
        fh.close()

        ask = QtWidgets.QMessageBox.question(
            self, self.tr("Reset driver..."), self.tr(
                "Reset piControl driver after successful uploading new piCtory "
                "configuration?\nThe process image will be interrupted for a "
                "short time!"
            ), QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No | QtWidgets.QMessageBox.Cancel
        )
        if ask == QtWidgets.QMessageBox.Cancel:
            return

        ec = helper.cm.call_remote_function(
            "set_pictoryrsc", file_binary, ask == QtWidgets.QMessageBox.Yes
        )

        if ec is None:
            QtWidgets.QMessageBox.critical(
                self, self.tr("Error"), self.tr(
                    "Got an network error while send data to Revolution Pi.\n"
                    "Please try again."
                )
            )
        elif ec == 0:
            helper.cm.settings.last_pictory_file = filename
            if ask == QtWidgets.QMessageBox.Yes:
                QtWidgets.QMessageBox.information(
                    self, self.tr("Success"), self.tr(
                        "The transfer of the piCtory configuration "
                        "and the reset of piControl have been "
                        "successfully executed."
                    ),
                )
            else:
                QtWidgets.QMessageBox.information(
                    self, self.tr("Success"), self.tr(
                        "The piCtory configuration was successfully transferred."
                    )
                )

        elif ec == -1:
            QtWidgets.QMessageBox.critical(
                self, self.tr("Error"), self.tr(
                    "Can not process the transferred file."
                )
            )
        elif ec == -2:
            QtWidgets.QMessageBox.critical(
                self, self.tr("Error"), self.tr(
                    "Can not find main elements in piCtory file."
                )
            )
        elif ec == -4:
            QtWidgets.QMessageBox.critical(
                self, self.tr("Error"), self.tr(
                    "Contained devices could not be found on Revolution "
                    "Pi. The configuration may be from a newer piCtory version!"
                )
            )
        elif ec == -5:
            QtWidgets.QMessageBox.critical(
                self, self.tr("Error"), self.tr(
                    "Could not load RAP catalog on Revolution Pi."
                )
            )
        elif ec < 0:
            QtWidgets.QMessageBox.critical(
                self, self.tr("Error"), self.tr(
                    "The piCtory configuration could not be "
                    "written on the Revolution Pi."
                )
            )
        elif ec > 0:
            QtWidgets.QMessageBox.warning(
                self, self.tr("Warning"), self.tr(
                    "The piCtroy configuration has been saved successfully.\n"
                    "An error occurred on piControl reset!"
                )
            )

    def check_replacedir(self, rootdir: str):
        """
        Return root dir of a extracted archive an piCtory configuration.

        This function checks the root dir of a extracted directory. It
        will check the existence of a piCtory configuration will will
        return that information in the return tuple.

        :param rootdir: Dir to check
        :return: Tuple with corrected root dir and piCtory configuration
        """
        lst_dir = os.listdir(rootdir)
        if len(lst_dir) == 1 and \
                os.path.isdir(os.path.join(rootdir, lst_dir[0])):
            return os.path.join(rootdir, lst_dir[0]), None

        if len(lst_dir) == 2:
            rscfile = None
            for fname in lst_dir:
                if fname.find(".rsc"):
                    rscfile = os.path.join(rootdir, fname)
            return os.path.join(rootdir, lst_dir[0]), rscfile

        else:
            return rootdir, None

    def create_filelist(self, rootdir):
        """Create a file list of a directory.

        :param rootdir: Root dir to crate file list
        :return: File list with <class 'str'>
        """
        filelist = []
        for tup_dir in os.walk(rootdir):
            for fname in tup_dir[2]:
                filelist.append(os.path.join(tup_dir[0], fname))
        return filelist

    def recover_gui(self):
        """Load saved GUI states."""
        self.cbb_format.setCurrentIndex(helper.settings.value("program/cbb_format_index", 0, int))
        self.cbx_pictory.setChecked(helper.settings.value("program/cbx_pictory_checked", False, bool))
        self.cbx_clear.setChecked(helper.settings.value("program/cbx_clear_checked", False, bool))

    @QtCore.pyqtSlot(int)
    def on_cbb_format_currentIndexChanged(self, index: int):
        helper.settings.setValue("program/cbb_format_index", index)
        self.cbx_pictory.setEnabled(index >= 1)

    @QtCore.pyqtSlot()
    def on_btn_program_download_clicked(self):
        """Download plc program from Revolution Pi."""
        if not helper.cm.connected:
            return

        if self.cbb_format.currentIndex() == 0:
            # Save files as zip archive
            diag_save = QtWidgets.QFileDialog(
                self, self.tr("Save ZIP archive..."),
                helper.cm.settings.last_zip_file or "{0}.zip".format(helper.cm.settings.name),
                self.tr("ZIP archive (*.zip);;All files (*.*)")
            )
            diag_save.setAcceptMode(QtWidgets.QFileDialog.AcceptSave)
            diag_save.setDefaultSuffix("zip")
            self.rejected.connect(diag_save.reject)

            if diag_save.exec() != QtWidgets.QFileDialog.AcceptSave or len(diag_save.selectedFiles()) != 1:
                return
            filename = diag_save.selectedFiles()[0]
            fh = open(filename, "wb")

            helper.cm.settings.last_zip_file = filename

        elif self.cbb_format.currentIndex() == 1:
            # Save files as TarGz archive
            diag_save = QtWidgets.QFileDialog(
                self, self.tr("Save TGZ archive..."),
                helper.cm.settings.last_tar_file or "{0}.tgz".format(helper.cm.settings.name),
                self.tr("TGZ archive (*.tgz);;All files (*.*)")
            )
            diag_save.setAcceptMode(QtWidgets.QFileDialog.AcceptSave)
            diag_save.setDefaultSuffix("tgz")
            self.rejected.connect(diag_save.reject)

            if diag_save.exec() != QtWidgets.QFileDialog.AcceptSave or len(diag_save.selectedFiles()) != 1:
                return
            filename = diag_save.selectedFiles()[0]
            fh = open(filename, "wb")

            helper.cm.settings.last_tar_file = filename

        else:
            # Other indexes are not allowed for download
            return

        plcfile = helper.cm.call_remote_function(
            "plcdownload",
            "zip" if self.cbb_format.currentIndex() == 0 else "tar",
            self.cbx_pictory.isChecked()
        )

        if plcfile is None:
            QtWidgets.QMessageBox.critical(
                self, self.tr("Error"), self.tr(
                    "Could not load PLC program from Revolution Pi."
                )
            )

        else:
            try:
                fh.write(plcfile.data)
                fh.close()

            except Exception as e:
                pi.logger.error(e)
                QtWidgets.QMessageBox.critical(
                    self, self.tr("Error"), self.tr(
                        "Coud not save the archive or extract the files!\n"
                        "Please retry.")
                )
            else:
                QtWidgets.QMessageBox.information(
                    self, self.tr("Success"), self.tr(
                        "Transfer successfully completed."
                    )
                )

    @QtCore.pyqtSlot()
    def on_btn_program_upload_clicked(self):
        """Upload plc program to Revolution Pi."""
        if not helper.cm.connected:
            return

        dirtmp = None

        def remove_temp():
            # Remove temp dir
            if dirtmp is not None:
                rmtree(dirtmp)

        dirselect = ""
        lst_files = []
        folder_name = ""
        rscfile = None

        if self.cbb_format.currentIndex() == 0:
            # Upload zip archive content
            diag_open = QtWidgets.QFileDialog(
                self, self.tr("Upload content of ZIP archive..."),
                helper.cm.settings.last_file_upload,
                self.tr("ZIP archive (*.zip);;All files (*.*)")
            )
            diag_open.setAcceptMode(QtWidgets.QFileDialog.AcceptOpen)
            diag_open.setFileMode(QtWidgets.QFileDialog.ExistingFile)
            diag_open.setDefaultSuffix("zip")
            self.rejected.connect(diag_open.reject)

            if diag_open.exec() != QtWidgets.QFileDialog.AcceptSave or len(diag_open.selectedFiles()) != 1:
                return

            filename = diag_open.selectedFiles()[0]
            helper.cm.settings.last_file_upload = filename
            if zipfile.is_zipfile(filename):
                dirtmp = mkdtemp()
                fhz = zipfile.ZipFile(filename)
                fhz.extractall(dirtmp)
                fhz.close()

                lst_files = self.create_filelist(dirtmp)
                dirselect, rscfile = self.check_replacedir(dirtmp)

            else:
                QtWidgets.QMessageBox.critical(
                    self, self.tr("Error"), self.tr(
                        "The selected file ist not a ZIP archive."
                    )
                )
                return

        elif self.cbb_format.currentIndex() == 1:
            # Upload TarGz content
            diag_open = QtWidgets.QFileDialog(
                self, self.tr("Upload content of TAR archive..."),
                helper.cm.settings.last_file_upload,
                self.tr("TAR archive (*.tgz);;All files (*.*)")
            )
            diag_open.setAcceptMode(QtWidgets.QFileDialog.AcceptOpen)
            diag_open.setFileMode(QtWidgets.QFileDialog.ExistingFile)
            diag_open.setDefaultSuffix("tgz")
            self.rejected.connect(diag_open.reject)

            if diag_open.exec() != QtWidgets.QFileDialog.AcceptSave or len(diag_open.selectedFiles()) != 1:
                return

            filename = diag_open.selectedFiles()[0]
            helper.cm.settings.last_file_upload = filename
            if tarfile.is_tarfile(filename):
                dirtmp = mkdtemp()
                fht = tarfile.open(filename)
                fht.extractall(dirtmp)
                fht.close()

                lst_files = self.create_filelist(dirtmp)
                dirselect, rscfile = self.check_replacedir(dirtmp)

            else:
                QtWidgets.QMessageBox.critical(
                    self, self.tr("Error"), self.tr(
                        "The selected file ist not a TAR archive."
                    )
                )
                return

        # No files selected
        if len(lst_files) == 0:
            QtWidgets.QMessageBox.warning(
                self, self.tr("No files to upload..."), self.tr(
                    "Found no files to upload in given location or archive."
                )
            )
            remove_temp()
            return

        # Clean up directory before upload
        clean_revpi = helper.cm.call_remote_function("plcuploadclean", default_value=False)
        if self.cbx_clear.isChecked() and not clean_revpi:
            QtWidgets.QMessageBox.critical(
                self, self.tr("Error"), self.tr(
                    "There was an error deleting the files on the Revolution Pi.\n"
                    "Upload aborted! Please try again."
                )
            )
            remove_temp()
            return

        plc_program_not_in_upload = True
        ec = 0

        for file_name in lst_files:

            if file_name == rscfile:
                # Do not send piCtory configuration
                continue

            # fixme: Fehlerabfang bei Dateilesen
            with open(file_name, "rb") as fh:

                # Generate relative file name for transfer
                if dirselect == "":
                    sendname = os.path.basename(file_name)
                else:
                    # Append folder name on complete folder transfer to crate it on Revolution Pi
                    sendname = os.path.join(
                        folder_name,
                        file_name.replace(dirselect, "")[1:]
                    )

                # Try to find given plc start program in upload files
                if sendname == self.dc.get("plcprogram", ""):
                    plc_program_not_in_upload = False

                ustatus = helper.cm.call_remote_function(
                    "plcupload", Binary(gzip.compress(fh.read())), sendname
                )
                if ustatus is None:
                    ec = -2
                    break
                elif not ustatus:
                    ec = -1
                    break

        if ec == 0:
            QtWidgets.QMessageBox.information(
                self, self.tr("Success"), self.tr(
                    "The PLC program was transferred successfully."
                )
            )
            self.lst_files = helper.cm.call_remote_function("get_filelist", default_value=[])
            self._load_settings(files_only=True)
            if plc_program_not_in_upload:
                QtWidgets.QMessageBox.warning(
                    self, self.tr("Information"), self.tr(
                        "Could not find the selected PLC start program in "
                        "uploaded files.\nThis is not an error, if the file "
                        "was already on the Revolution Pi. Check PLC start "
                        "program field"
                    )
                )

            if self.cbx_pictory.isChecked():
                if rscfile is not None:
                    self._upload_pictory(rscfile)
                else:
                    QtWidgets.QMessageBox.critical(
                        self, self.tr("Error"), self.tr(
                            "There is no piCtory configuration in this archive."
                        )
                    )

        elif ec == -1:
            QtWidgets.QMessageBox.critical(
                self, self.tr("Error"), self.tr(
                    "The Revolution Pi could not process some parts of the transmission."
                )
            )

        elif ec == -2:
            QtWidgets.QMessageBox.critical(
                self, self.tr("Error"), self.tr(
                    "Errors occurred during transmission."
                )
            )

        remove_temp()

    # endregion # # # # #

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    # region #      REGION: Control files

    @QtCore.pyqtSlot()
    def on_btn_pictory_download_clicked(self):
        """Download piCtory configuration."""
        if not helper.cm.connected:
            return

        diag_save = QtWidgets.QFileDialog(
            self, self.tr("Save piCtory file..."),
            os.path.join(
                helper.cm.settings.last_dir_pictory,
                "{0}.rsc".format(helper.cm.settings.name)
            ),
            self.tr("piCtory file (*.rsc);;All files (*.*)")
        )
        diag_save.setAcceptMode(QtWidgets.QFileDialog.AcceptSave)
        diag_save.setDefaultSuffix("rsc")
        self.rejected.connect(diag_save.reject)

        if diag_save.exec() != QtWidgets.QFileDialog.AcceptSave or len(diag_save.selectedFiles()) != 1:
            return

        filename = diag_save.selectedFiles()[0]
        helper.cm.settings.last_dir_pictory = os.path.dirname(filename)
        bin_buffer = helper.cm.call_remote_function("get_pictoryrsc")  # type: Binary
        if bin_buffer is None:
            QtWidgets.QMessageBox.critical(
                self, self.tr("Error"), self.tr(
                    "Could not load piCtory file from Revolution Pi."
                )
            )
        else:
            fh = open(filename, "wb")
            fh.write(bin_buffer.data)
            fh.close()

            QtWidgets.QMessageBox.information(
                self, self.tr("Success"), self.tr(
                    "piCtory configuration successfully loaded and saved to:\n{0}."
                ).format(filename)
            )

    @QtCore.pyqtSlot()
    def on_btn_pictory_upload_clicked(self):
        if not helper.cm.connected:
            return

        diag_open = QtWidgets.QFileDialog(
            self, self.tr("Upload piCtory file..."),
            helper.cm.settings.last_pictory_file or "{0}.rsc".format(helper.cm.settings.name),
            self.tr("piCtory file (*.rsc);;All files (*.*)")
        )
        diag_open.setAcceptMode(QtWidgets.QFileDialog.AcceptOpen)
        diag_open.setFileMode(QtWidgets.QFileDialog.ExistingFile)
        diag_open.setDefaultSuffix("rsc")
        self.rejected.connect(diag_open.reject)

        if diag_open.exec() != QtWidgets.QFileDialog.AcceptSave or len(diag_open.selectedFiles()) != 1:
            return

        self._upload_pictory(diag_open.selectedFiles()[0])

    @QtCore.pyqtSlot()
    def on_btn_procimg_download_clicked(self):
        """Download process image."""
        if not helper.cm.connected:
            return

        diag_save = QtWidgets.QFileDialog(
            self,
            self.tr("Save piControl file..."),
            os.path.join(
                helper.cm.settings.last_dir_picontrol,
                "{0}.img".format(helper.cm.settings.name)
            ),
            self.tr("Process image file (*.img);;All files (*.*)")
        )
        diag_save.setAcceptMode(QtWidgets.QFileDialog.AcceptSave)
        diag_save.setDefaultSuffix("img")
        self.rejected.connect(diag_save.reject)

        if diag_save.exec() != QtWidgets.QFileDialog.AcceptSave or len(diag_save.selectedFiles()) != 1:
            return

        filename = diag_save.selectedFiles()[0]
        helper.cm.settings.last_dir_picontrol = os.path.dirname(filename)
        bin_buffer = helper.cm.call_remote_function("get_procimg")  # type: Binary

        if bin_buffer is None:
            QtWidgets.QMessageBox.critical(
                self, self.tr("Error"), self.tr(
                    "Could not load process image from Revolution Pi."
                )
            )
        else:
            fh = open(filename, "wb")
            fh.write(bin_buffer.data)
            fh.close()

            QtWidgets.QMessageBox.information(
                self, self.tr("Success"), self.tr(
                    "Process image successfully loaded and saved to:\n{0}."
                ).format(filename)
            )

    # endregion # # # # #



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'revpiplclist.ui'
#
# Created by: PyQt5 UI code generator 5.15.7
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_diag_connections(object):
    def setupUi(self, diag_connections):
        diag_connections.setObjectName("diag_connections")
        diag_connections.resize(496, 569)
        self.gridLayout = QtWidgets.QGridLayout(diag_connections)
        self.gridLayout.setObjectName("gridLayout")
        self.tab_properties = QtWidgets.QTabWidget(diag_connections)
        self.tab_properties.setObjectName("tab_properties")
        self.tab_connection = QtWidgets.QWidget()
        self.tab_connection.setObjectName("tab_connection")
        self.formLayout_2 = QtWidgets.QFormLayout(self.tab_connection)
        self.formLayout_2.setObjectName("formLayout_2")
        self.lbl_name = QtWidgets.QLabel(self.tab_connection)
        self.lbl_name.setObjectName("lbl_name")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.lbl_name)
        self.txt_name = QtWidgets.QLineEdit(self.tab_connection)
        self.txt_name.setObjectName("txt_name")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.txt_name)
        self.lbl_address = QtWidgets.QLabel(self.tab_connection)
        self.lbl_address.setObjectName("lbl_address")
        self.formLayout_2.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.lbl_address)
        self.txt_address = QtWidgets.QLineEdit(self.tab_connection)
        self.txt_address.setObjectName("txt_address")
        self.formLayout_2.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.txt_address)
        self.lbl_port = QtWidgets.QLabel(self.tab_connection)
        self.lbl_port.setObjectName("lbl_port")
        self.formLayout_2.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.lbl_port)
        self.sbx_port = QtWidgets.QSpinBox(self.tab_connection)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.sbx_port.sizePolicy().hasHeightForWidth())
        self.sbx_port.setSizePolicy(sizePolicy)
        self.sbx_port.setMinimum(1)
        self.sbx_port.setMaximum(65535)
        self.sbx_port.setProperty("value", 55123)
        self.sbx_port.setObjectName("sbx_port")
        self.formLayout_2.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.sbx_port)
        self.lbl_timeout = QtWidgets.QLabel(self.tab_connection)
        self.lbl_timeout.setObjectName("lbl_timeout")
        self.formLayout_2.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.lbl_timeout)
        self.sbx_timeout = QtWidgets.QSpinBox(self.tab_connection)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.sbx_timeout.sizePolicy().hasHeightForWidth())
        self.sbx_timeout.setSizePolicy(sizePolicy)
        self.sbx_timeout.setMinimum(5)
        self.sbx_timeout.setMaximum(30)
        self.sbx_timeout.setObjectName("sbx_timeout")
        self.formLayout_2.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.sbx_timeout)
        self.lbl_folder = QtWidgets.QLabel(self.tab_connection)
        self.lbl_folder.setObjectName("lbl_folder")
        self.formLayout_2.setWidget(4, QtWidgets.QFormLayout.LabelRole, self.lbl_folder)
        self.cbb_folder = QtWidgets.QComboBox(self.tab_connection)
        self.cbb_folder.setEditable(True)
        self.cbb_folder.setSizeAdjustPolicy(QtWidgets.QComboBox.AdjustToContents)
        self.cbb_folder.setObjectName("cbb_folder")
        self.cbb_folder.addItem("")
        self.cbb_folder.setItemText(0, "")
        self.formLayout_2.setWidget(4, QtWidgets.QFormLayout.FieldRole, self.cbb_folder)
        self.tab_properties.addTab(self.tab_connection, "")
        self.tab_ssh = QtWidgets.QWidget()
        self.tab_ssh.setObjectName("tab_ssh")
        self.formLayout = QtWidgets.QFormLayout(self.tab_ssh)
        self.formLayout.setObjectName("formLayout")
        self.lbl_ssh_use_tunnel = QtWidgets.QLabel(self.tab_ssh)
        self.lbl_ssh_use_tunnel.setObjectName("lbl_ssh_use_tunnel")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.lbl_ssh_use_tunnel)
        self.cbx_ssh_use_tunnel = QtWidgets.QCheckBox(self.tab_ssh)
        self.cbx_ssh_use_tunnel.setText("")
        self.cbx_ssh_use_tunnel.setChecked(True)
        self.cbx_ssh_use_tunnel.setObjectName("cbx_ssh_use_tunnel")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.cbx_ssh_use_tunnel)
        self.lbl_ssh_port = QtWidgets.QLabel(self.tab_ssh)
        self.lbl_ssh_port.setObjectName("lbl_ssh_port")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.lbl_ssh_port)
        self.sbx_ssh_port = QtWidgets.QSpinBox(self.tab_ssh)
        self.sbx_ssh_port.setMaximum(65535)
        self.sbx_ssh_port.setProperty("value", 22)
        self.sbx_ssh_port.setObjectName("sbx_ssh_port")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.sbx_ssh_port)
        self.lbl_ssh_user = QtWidgets.QLabel(self.tab_ssh)
        self.lbl_ssh_user.setObjectName("lbl_ssh_user")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.lbl_ssh_user)
        self.txt_ssh_user = QtWidgets.QLineEdit(self.tab_ssh)
        self.txt_ssh_user.setText("pi")
        self.txt_ssh_user.setObjectName("txt_ssh_user")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.txt_ssh_user)
        self.tab_properties.addTab(self.tab_ssh, "")
        self.gridLayout.addWidget(self.tab_properties, 1, 0, 1, 2)
        self.btn_box = QtWidgets.QDialogButtonBox(diag_connections)
        self.btn_box.setOrientation(QtCore.Qt.Horizontal)
        self.btn_box.setStandardButtons(QtWidgets.QDialogButtonBox.Discard|QtWidgets.QDialogButtonBox.Save)
        self.btn_box.setObjectName("btn_box")
        self.gridLayout.addWidget(self.btn_box, 2, 0, 1, 2)
        self.tre_connections = QtWidgets.QTreeWidget(diag_connections)
        self.tre_connections.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.tre_connections.setObjectName("tre_connections")
        self.gridLayout.addWidget(self.tre_connections, 0, 0, 1, 1)
        self.vl_edit = QtWidgets.QVBoxLayout()
        self.vl_edit.setObjectName("vl_edit")
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.vl_edit.addItem(spacerItem)
        self.btn_up = QtWidgets.QPushButton(diag_connections)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/action/ico/arrow-up.ico"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.btn_up.setIcon(icon)
        self.btn_up.setObjectName("btn_up")
        self.vl_edit.addWidget(self.btn_up)
        self.btn_down = QtWidgets.QPushButton(diag_connections)
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/action/ico/arrow-down.ico"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.btn_down.setIcon(icon1)
        self.btn_down.setObjectName("btn_down")
        self.vl_edit.addWidget(self.btn_down)
        self.btn_delete = QtWidgets.QPushButton(diag_connections)
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(":/action/ico/edit-delete-6.ico"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.btn_delete.setIcon(icon2)
        self.btn_delete.setObjectName("btn_delete")
        self.vl_edit.addWidget(self.btn_delete)
        self.btn_add_dir = QtWidgets.QPushButton(diag_connections)
        icon3 = QtGui.QIcon()
        icon3.addPixmap(QtGui.QPixmap(":/action/ico/folder-open.ico"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.btn_add_dir.setIcon(icon3)
        self.btn_add_dir.setObjectName("btn_add_dir")
        self.vl_edit.addWidget(self.btn_add_dir)
        self.btn_add = QtWidgets.QPushButton(diag_connections)
        icon4 = QtGui.QIcon()
        icon4.addPixmap(QtGui.QPixmap(":/action/ico/edit-add.ico"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.btn_add.setIcon(icon4)
        self.btn_add.setObjectName("btn_add")
        self.vl_edit.addWidget(self.btn_add)
        self.gridLayout.addLayout(self.vl_edit, 0, 1, 1, 1)

        self.retranslateUi(diag_connections)
        self.tab_properties.setCurrentIndex(0)
        self.btn_box.rejected.connect(diag_connections.reject) # type: ignore
        self.btn_box.accepted.connect(diag_connections.accept) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(diag_connections)

    def retranslateUi(self, diag_connections):
        _translate = QtCore.QCoreApplication.translate
        diag_connections.setWindowTitle(_translate("diag_connections", "Revolution Pi connections"))
        self.lbl_name.setText(_translate("diag_connections", "Display name:"))
        self.lbl_address.setText(_translate("diag_connections", "Address (DNS/IP):"))
        self.lbl_port.setText(_translate("diag_connections", "Port (Default {0}):"))
        self.lbl_timeout.setText(_translate("diag_connections", "Connection timeout:"))
        self.sbx_timeout.setSuffix(_translate("diag_connections", " sec."))
        self.lbl_folder.setText(_translate("diag_connections", "Sub folder:"))
        self.tab_properties.setTabText(self.tab_properties.indexOf(self.tab_connection), _translate("diag_connections", "Connection"))
        self.lbl_ssh_use_tunnel.setText(_translate("diag_connections", "Connect over SSH tunnel:"))
        self.lbl_ssh_port.setText(_translate("diag_connections", "SSH port:"))
        self.lbl_ssh_user.setText(_translate("diag_connections", "SSH user name:"))
        self.tab_properties.setTabText(self.tab_properties.indexOf(self.tab_ssh), _translate("diag_connections", "Over SSH"))
        self.tre_connections.headerItem().setText(0, _translate("diag_connections", "Connection name"))
        self.tre_connections.headerItem().setText(1, _translate("diag_connections", "Address"))
from . import ressources_rc


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    diag_connections = QtWidgets.QDialog()
    ui = Ui_diag_connections()
    ui.setupUi(diag_connections)
    diag_connections.show()
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-
"""Program information of local an remote system."""
__author__ = "Sven Sager"
__copyright__ = "Copyright (C) 2023 Sven Sager"
__license__ = "GPLv2"

from PyQt5 import QtCore, QtGui, QtWidgets

from . import helper
from .ui.revpiinfo_ui import Ui_diag_revpiinfo


class RevPiInfo(QtWidgets.QDialog, Ui_diag_revpiinfo):
    """Version information window."""

    def __init__(self, version: str, parent=None):
        super().__init__(parent)
        self.setupUi(self)

        self._debug_load = False

        self.lbl_version_control.setText(version)
        self.lbl_version_control.mousePressEvent = self.lbl_version_mousePressEvent

    def exec(self) -> int:
        self.lbl_version_pyload.setText(
            helper.cm.pyload_version_str
            if helper.cm.connected else "-"
        )
        self._load_lst_files()
        return super().exec()

    def lbl_version_mousePressEvent(self, a0: QtGui.QMouseEvent):
        if a0.button() == QtCore.Qt.MidButton:
            self._debug_load = not self._debug_load
            self._load_lst_files()

    def _load_lst_files(self):
        """Load files from working dir on Revolution Pi."""
        self.lst_files.clear()

        if self._debug_load:
            lst = helper.cm.xml_funcs
        elif helper.cm.connected:
            lst = helper.cm.call_remote_function(
                "get_filelist",
                default_value=[self.tr("Can not load file list")]
            )
        else:
            lst = [self.tr("Not connected")]

        lst.sort()
        self.lst_files.insertItems(0, lst)



# -*- coding: utf-8 -*-
"""Simulator for piControl."""
__author__ = "Sven Sager"
__copyright__ = "Copyright (C) 2023 Sven Sager"
__license__ = "GPLv2"

from os import W_OK, access
from os.path import basename, dirname, exists, join

from PyQt5 import QtCore, QtGui, QtWidgets

from . import helper
from .ui.simulator_ui import Ui_diag_simulator


class Simulator(QtWidgets.QDialog, Ui_diag_simulator):
    """
    This is a configuration dialog for the simulator of piControl. The
    selected values will be saved in QSettings section 'simulator' and can be
    accessed by simulator starting classes.
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setupUi(self)
        self.clean_procimg = False
        self.max_items = 5

        self.cbb_history.addItem("", "")
        lst_configrsc = helper.settings.value("simulator/history_configrsc", [], list)
        lst_procimg = helper.settings.value("simulator/history_procimg", [], list)
        for i in range(len(lst_configrsc)):
            self.cbb_history.addItem(lst_configrsc[i], lst_procimg[i])

        self.cbx_stop_remove.setChecked(helper.settings.value("simulator/stop_remove", False, bool))
        self.rb_restart_pictory.setChecked(helper.settings.value("simulator/restart_pictory", False, bool))
        self.rb_restart_zero.setChecked(helper.settings.value("simulator/restart_zero", False, bool))

        self.btn_start_pictory.setEnabled(False)
        self.btn_start_empty.setEnabled(False)
        self.btn_start_nochange.setEnabled(False)

        self.txt_configrsc.textChanged.connect(self.on_txt_textChanged)
        self.txt_procimg.textChanged.connect(self.on_txt_textChanged)

    def _save_gui(self) -> None:
        helper.settings.setValue("simulator/stop_remove", self.cbx_stop_remove.isChecked())
        helper.settings.setValue("simulator/restart_pictory", self.rb_restart_pictory.isChecked())
        helper.settings.setValue("simulator/restart_zero", self.rb_restart_zero.isChecked())

    def accept(self) -> None:
        self.cbb_history.removeItem(0)
        if self.cbb_history.findText(self.txt_configrsc.text()) == -1:
            self.cbb_history.addItem(self.txt_configrsc.text(), self.txt_procimg.text())
        if self.cbb_history.count() > self.max_items:
            self.cbb_history.removeItem(self.max_items)

        helper.settings.setValue("simulator/configrsc", self.txt_configrsc.text())
        helper.settings.setValue("simulator/procimg", self.txt_procimg.text())
        self._save_gui()

        lst_configrsc = []
        lst_procimg = []
        for i in range(self.cbb_history.count()):
            lst_configrsc.append(self.cbb_history.itemText(i))
            lst_procimg.append(self.cbb_history.itemData(i))
        helper.settings.setValue("simulator/history_configrsc", lst_configrsc)
        helper.settings.setValue("simulator/history_procimg", lst_procimg)

        self.clean_procimg = self.sender() is self.btn_start_empty

        super().accept()

    def closeEvent(self, a0: QtGui.QCloseEvent) -> None:
        self._save_gui()

    @QtCore.pyqtSlot()
    def on_btn_configrsc_clicked(self) -> None:
        diag_open = QtWidgets.QFileDialog(
            self, self.tr("Select downloaded piCtory file..."),
            helper.settings.value("simulator/last_dir", ".", str),
            self.tr("piCtory file (*.rsc);;All files (*.*)")
        )
        diag_open.setAcceptMode(QtWidgets.QFileDialog.AcceptOpen)
        diag_open.setFileMode(QtWidgets.QFileDialog.ExistingFile)
        diag_open.setDefaultSuffix("rsc")

        if diag_open.exec() != QtWidgets.QFileDialog.AcceptSave or len(diag_open.selectedFiles()) != 1:
            diag_open.deleteLater()
            return

        configrsc_file = diag_open.selectedFiles()[0]
        dir_name = dirname(configrsc_file)
        procimg_file = join(dir_name, "{0}.img".format(basename(configrsc_file).rsplit(".", maxsplit=1)[0]))
        self.txt_configrsc.setText(configrsc_file)
        self.txt_procimg.setText(procimg_file)

        helper.settings.setValue("simulator/last_dir", dir_name)
        diag_open.deleteLater()

    @QtCore.pyqtSlot(int)
    def on_cbb_history_currentIndexChanged(self, index: int) -> None:
        if index == 0:
            return
        self.txt_configrsc.setText(self.cbb_history.itemText(index))
        self.txt_procimg.setText(self.cbb_history.itemData(index))

    @QtCore.pyqtSlot(str)
    def on_txt_textChanged(self, text: str) -> None:
        configrsc_file = self.txt_configrsc.text()
        procimg_file = self.txt_procimg.text()
        if configrsc_file and procimg_file:
            file_access = access(procimg_file, W_OK) if exists(procimg_file) else access(dirname(procimg_file), W_OK)
            self.txt_info.setPlainText("configrsc=\"{0}\", procimg=\"{1}\"".format(configrsc_file, procimg_file))
            self.btn_start_pictory.setEnabled(file_access)
            self.btn_start_empty.setEnabled(file_access)
            self.btn_start_nochange.setEnabled(file_access and exists(procimg_file))
        else:
            self.txt_info.clear()
            self.btn_start_pictory.setEnabled(False)
            self.btn_start_empty.setEnabled(False)
            self.btn_start_nochange.setEnabled(False)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'files.ui'
#
# Created by: PyQt5 UI code generator 5.15.7
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_win_files(object):
    def setupUi(self, win_files):
        win_files.setObjectName("win_files")
        win_files.resize(725, 519)
        self.centralwidget = QtWidgets.QWidget(win_files)
        self.centralwidget.setObjectName("centralwidget")
        self.gridLayout = QtWidgets.QGridLayout(self.centralwidget)
        self.gridLayout.setObjectName("gridLayout")
        self.splitter = QtWidgets.QSplitter(self.centralwidget)
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.splitter.setChildrenCollapsible(False)
        self.splitter.setObjectName("splitter")
        self.verticalLayoutWidget = QtWidgets.QWidget(self.splitter)
        self.verticalLayoutWidget.setObjectName("verticalLayoutWidget")
        self.vl_local = QtWidgets.QVBoxLayout(self.verticalLayoutWidget)
        self.vl_local.setContentsMargins(0, 0, 0, 0)
        self.vl_local.setObjectName("vl_local")
        self.gb_select_local = QtWidgets.QGroupBox(self.verticalLayoutWidget)
        self.gb_select_local.setObjectName("gb_select_local")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.gb_select_local)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.lbl_select_local = QtWidgets.QLabel(self.gb_select_local)
        self.lbl_select_local.setObjectName("lbl_select_local")
        self.gridLayout_2.addWidget(self.lbl_select_local, 0, 0, 1, 1)
        self.btn_select_local = QtWidgets.QPushButton(self.gb_select_local)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/action/ico/folder-open.ico"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.btn_select_local.setIcon(icon)
        self.btn_select_local.setIconSize(QtCore.QSize(24, 24))
        self.btn_select_local.setAutoDefault(False)
        self.btn_select_local.setObjectName("btn_select_local")
        self.gridLayout_2.addWidget(self.btn_select_local, 0, 1, 1, 1)
        self.btn_refresh_local = QtWidgets.QPushButton(self.gb_select_local)
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/action/ico/refresh.ico"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.btn_refresh_local.setIcon(icon1)
        self.btn_refresh_local.setIconSize(QtCore.QSize(24, 24))
        self.btn_refresh_local.setObjectName("btn_refresh_local")
        self.gridLayout_2.addWidget(self.btn_refresh_local, 0, 2, 1, 1)
        self.lbl_path_local = QtWidgets.QLabel(self.gb_select_local)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lbl_path_local.sizePolicy().hasHeightForWidth())
        self.lbl_path_local.setSizePolicy(sizePolicy)
        self.lbl_path_local.setToolTip("/")
        self.lbl_path_local.setText("/")
        self.lbl_path_local.setObjectName("lbl_path_local")
        self.gridLayout_2.addWidget(self.lbl_path_local, 1, 0, 1, 3)
        self.gridLayout_2.setColumnStretch(0, 1)
        self.vl_local.addWidget(self.gb_select_local)
        self.hl_revpi_2 = QtWidgets.QHBoxLayout()
        self.hl_revpi_2.setObjectName("hl_revpi_2")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.hl_revpi_2.addItem(spacerItem)
        self.btn_to_right = QtWidgets.QPushButton(self.verticalLayoutWidget)
        self.btn_to_right.setText("")
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(":/action/ico/arrow-right.ico"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.btn_to_right.setIcon(icon2)
        self.btn_to_right.setIconSize(QtCore.QSize(24, 24))
        self.btn_to_right.setAutoDefault(False)
        self.btn_to_right.setObjectName("btn_to_right")
        self.hl_revpi_2.addWidget(self.btn_to_right)
        self.vl_local.addLayout(self.hl_revpi_2)
        self.tree_files_local = QtWidgets.QTreeWidget(self.verticalLayoutWidget)
        self.tree_files_local.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.tree_files_local.setSelectionMode(QtWidgets.QAbstractItemView.MultiSelection)
        self.tree_files_local.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.tree_files_local.setIconSize(QtCore.QSize(24, 24))
        self.tree_files_local.setObjectName("tree_files_local")
        self.tree_files_local.headerItem().setText(0, "1")
        self.tree_files_local.header().setVisible(False)
        self.vl_local.addWidget(self.tree_files_local)
        self.gridLayoutWidget_2 = QtWidgets.QWidget(self.splitter)
        self.gridLayoutWidget_2.setObjectName("gridLayoutWidget_2")
        self.vl_revpi = QtWidgets.QVBoxLayout(self.gridLayoutWidget_2)
        self.vl_revpi.setContentsMargins(0, 0, 0, 0)
        self.vl_revpi.setObjectName("vl_revpi")
        self.gb_select_revpi = QtWidgets.QGroupBox(self.gridLayoutWidget_2)
        self.gb_select_revpi.setObjectName("gb_select_revpi")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.gb_select_revpi)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.lbl_path_revpi = QtWidgets.QLabel(self.gb_select_revpi)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lbl_path_revpi.sizePolicy().hasHeightForWidth())
        self.lbl_path_revpi.setSizePolicy(sizePolicy)
        self.lbl_path_revpi.setToolTip("/")
        self.lbl_path_revpi.setText("/")
        self.lbl_path_revpi.setObjectName("lbl_path_revpi")
        self.gridLayout_3.addWidget(self.lbl_path_revpi, 1, 0, 1, 2)
        self.lbl_select_revpi = QtWidgets.QLabel(self.gb_select_revpi)
        self.lbl_select_revpi.setObjectName("lbl_select_revpi")
        self.gridLayout_3.addWidget(self.lbl_select_revpi, 0, 0, 1, 1)
        self.btn_refresh_revpi = QtWidgets.QPushButton(self.gb_select_revpi)
        self.btn_refresh_revpi.setIcon(icon1)
        self.btn_refresh_revpi.setIconSize(QtCore.QSize(24, 24))
        self.btn_refresh_revpi.setObjectName("btn_refresh_revpi")
        self.gridLayout_3.addWidget(self.btn_refresh_revpi, 0, 1, 1, 1)
        self.gridLayout_3.setColumnStretch(0, 1)
        self.vl_revpi.addWidget(self.gb_select_revpi)
        self.hl_revpi = QtWidgets.QHBoxLayout()
        self.hl_revpi.setObjectName("hl_revpi")
        self.btn_to_left = QtWidgets.QPushButton(self.gridLayoutWidget_2)
        self.btn_to_left.setText("")
        icon3 = QtGui.QIcon()
        icon3.addPixmap(QtGui.QPixmap(":/action/ico/arrow-left.ico"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.btn_to_left.setIcon(icon3)
        self.btn_to_left.setIconSize(QtCore.QSize(24, 24))
        self.btn_to_left.setAutoDefault(False)
        self.btn_to_left.setObjectName("btn_to_left")
        self.hl_revpi.addWidget(self.btn_to_left)
        self.btn_delete_revpi = QtWidgets.QPushButton(self.gridLayoutWidget_2)
        self.btn_delete_revpi.setText("")
        icon4 = QtGui.QIcon()
        icon4.addPixmap(QtGui.QPixmap(":/action/ico/edit-delete.ico"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.btn_delete_revpi.setIcon(icon4)
        self.btn_delete_revpi.setIconSize(QtCore.QSize(24, 24))
        self.btn_delete_revpi.setAutoDefault(False)
        self.btn_delete_revpi.setObjectName("btn_delete_revpi")
        self.hl_revpi.addWidget(self.btn_delete_revpi)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.hl_revpi.addItem(spacerItem1)
        self.vl_revpi.addLayout(self.hl_revpi)
        self.tree_files_revpi = QtWidgets.QTreeWidget(self.gridLayoutWidget_2)
        self.tree_files_revpi.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.tree_files_revpi.setSelectionMode(QtWidgets.QAbstractItemView.MultiSelection)
        self.tree_files_revpi.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.tree_files_revpi.setIconSize(QtCore.QSize(24, 24))
        self.tree_files_revpi.setObjectName("tree_files_revpi")
        self.tree_files_revpi.headerItem().setText(0, "1")
        self.tree_files_revpi.header().setVisible(False)
        self.vl_revpi.addWidget(self.tree_files_revpi)
        self.gridLayout.addWidget(self.splitter, 0, 0, 1, 1)
        self.btn_all = QtWidgets.QPushButton(self.centralwidget)
        self.btn_all.setObjectName("btn_all")
        self.gridLayout.addWidget(self.btn_all, 1, 0, 1, 1)
        win_files.setCentralWidget(self.centralwidget)
        self.statusbar = QtWidgets.QStatusBar(win_files)
        self.statusbar.setObjectName("statusbar")
        win_files.setStatusBar(self.statusbar)

        self.retranslateUi(win_files)
        QtCore.QMetaObject.connectSlotsByName(win_files)

    def retranslateUi(self, win_files):
        _translate = QtCore.QCoreApplication.translate
        win_files.setWindowTitle(_translate("win_files", "File manager"))
        self.gb_select_local.setTitle(_translate("win_files", "Local computer"))
        self.lbl_select_local.setText(_translate("win_files", "Path to development root:"))
        self.btn_select_local.setToolTip(_translate("win_files", "Open developer root directory"))
        self.btn_refresh_local.setToolTip(_translate("win_files", "Reload file list"))
        self.tree_files_local.setSortingEnabled(True)
        self.gb_select_revpi.setTitle(_translate("win_files", "Revolution Pi"))
        self.lbl_select_revpi.setText(_translate("win_files", "RevPiPyLoad working directory:"))
        self.btn_refresh_revpi.setToolTip(_translate("win_files", "Reload file list"))
        self.tree_files_revpi.setSortingEnabled(True)
        self.btn_all.setText(_translate("win_files", "Stop - Upload - Start"))
from . import ressources_rc


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    win_files = QtWidgets.QMainWindow()
    ui = Ui_win_files()
    ui.setupUi(win_files)
    win_files.show()
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-
"""
Connect to a remote host and tunnel a port.

This was crated on base of the paramiko library demo file forward.py, see on
GitHub https://github.com/paramiko/paramiko/blob/main/demos/forward.py
"""
__author__ = "Sven Sager"
__copyright__ = "Copyright (C) 2023 Sven Sager"
__license__ = "GPLv2"

import select
from logging import getLogger
from socketserver import BaseRequestHandler, ThreadingTCPServer
from threading import Thread
from typing import Tuple, Union

from paramiko.client import MissingHostKeyPolicy, SSHClient
from paramiko.rsakey import RSAKey
from paramiko.ssh_exception import PasswordRequiredException
from paramiko.transport import Transport

log = getLogger("ssh_tunneling")


class ForwardServer(ThreadingTCPServer):
    daemon_threads = True
    allow_reuse_address = True


class Handler(BaseRequestHandler):
    def handle(self):
        try:
            chan = self.ssh_transport.open_channel(
                "direct-tcpip",
                (self.chain_host, self.chain_port),
                self.request.getpeername(),
            )
        except Exception as e:
            log.error(e)
            return
        if chan is None:
            log.error("Could not create a ssh channel")
            return

        log.info("Starting tunnel exchange loop")

        while True:
            r, w, x = select.select([self.request, chan], [], [], 5.0)
            if self.request in r:
                data = self.request.recv(1024)
                if len(data) == 0:
                    break
                chan.send(data)
            if chan in r:
                data = chan.recv(1024)
                if len(data) == 0:
                    break
                self.request.send(data)

        log.info("Stopped tunnel exchange loop")

        chan.close()
        self.request.close()


class SSHLocalTunnel:

    def __init__(self, remote_tunnel_port: int, ssh_host: str, ssh_port: int = 22):
        """
        Connect to a ssh remote host and tunnel a port to your host.

        :param remote_tunnel_port: Port on the remote host to tunnel through ssh
        :param ssh_host: ssh remote host address
        :param ssh_port: ssh remote host port
        """
        self._remote_tunnel_port = remote_tunnel_port
        self._ssh_host = ssh_host
        self._ssh_port = ssh_port

        self._th_server = Thread()

        self._ssh_client = SSHClient()
        self._ssh_client.set_missing_host_key_policy(MissingHostKeyPolicy())

        self._ssh_transport = None  # type: Transport
        self._forward_server = None  # type: ThreadingTCPServer
        self._local_tunnel_port = None  # type: int

    def __th_target(self):
        """Server thread for socket mirror."""
        self._forward_server.serve_forever()

    def _configure_forward_server(self) -> int:
        """
        Configure forward server for port mirror.

        :return: Local port on wich the remote port is connected
        """
        self._ssh_transport = self._ssh_client.get_transport()

        class SubHandler(Handler):
            chain_host = "127.0.0.1"
            chain_port = self._remote_tunnel_port
            ssh_transport = self._ssh_transport

        self._forward_server = ForwardServer(("127.0.0.1", 0), SubHandler)
        self._local_tunnel_port = self._forward_server.socket.getsockname()[1]

        self._th_server = Thread(target=self.__th_target)
        self._th_server.start()

        return self._local_tunnel_port

    def connect_by_credentials(self, username: str, password: str) -> int:
        """
        Connect to a ssh remote host and tunnel specified port of localhost.

        :return: Local port on wich the remote port is connected
        """
        if self._th_server.is_alive():
            raise RuntimeError("Already connected")

        self._ssh_client.connect(
            hostname=self._ssh_host,
            port=self._ssh_port,
            username=username,
            password=password,
        )

        return self._configure_forward_server()

    def connect_by_keyfile(self, username: str, key_file: str, key_password: str = None) -> int:
        """
        Connect to a ssh remote host and tunnel specified port of localhost.

        :return: Local port on wich the remote port is connected
        """
        if self._th_server.is_alive():
            raise RuntimeError("Already connected")

        if self.key_file_password_protected(key_file):
            private_key = RSAKey.from_private_key_file(key_file, key_password)
        else:
            private_key = RSAKey.from_private_key_file(key_file)

        self._ssh_client.connect(
            hostname=self._ssh_host,
            port=self._ssh_port,
            username=username,
            pkey=private_key,
            look_for_keys=True,
        )

        return self._configure_forward_server()

    def disconnect(self):
        """Close SSH tunnel connection."""
        self._local_tunnel_port = None
        if self._forward_server:
            self._forward_server.shutdown()
            self._forward_server.server_close()
        if self._ssh_transport:
            self._ssh_transport.close()
        self._ssh_client.close()

    @staticmethod
    def key_file_password_protected(key_file: str) -> bool:
        try:
            RSAKey.from_private_key_file(key_file)
        except PasswordRequiredException:
            return True
        return False

    def send_cmd(self, cmd: str, timeout: float = None) -> Union[Tuple[str, str], Tuple[None, None]]:
        """
        Send simple command to ssh host.

        The output of stdout and stderr is returned as a tuple of two elements.
        This elements could be None, in case of an internal error.

        :param cmd: Shell command to execute on remote host
        :param timeout: Timeout for execution
        :return: Tuple with stdout and stderr
        """
        if not self._th_server.is_alive():
            raise RuntimeError("Not connected")

        try:
            _, stdout, stderr = self._ssh_client.exec_command(cmd, 1024, timeout)
            buffer_out = stdout.read()
            buffer_err = stderr.read()

            return buffer_out.decode(), buffer_err.decode()
        except Exception as e:
            log.error(e)
            return None, None

    @property
    def connected(self):
        """Check connection state of ssh tunnel."""
        return self._ssh_transport and self._ssh_transport.is_active()

    @property
    def local_tunnel_port(self) -> int:
        return self._local_tunnel_port



# -*- coding: utf-8 -*-
"""Helper functions for this application."""
__author__ = "Sven Sager"
__copyright__ = "Copyright (C) 2023 Sven Sager"
__license__ = "GPLv2"

import pickle
import socket
from enum import IntEnum
from http.client import CannotSendRequest
from os import environ, remove
from os.path import exists
from queue import Queue
from re import search
from threading import Lock
from uuid import uuid4
from xmlrpc.client import Binary, ServerProxy

from PyQt5 import QtCore
from paramiko.ssh_exception import AuthenticationException

from . import proginit as pi
from .ssh_tunneling.server import SSHLocalTunnel

settings = QtCore.QSettings("revpimodio.org", "revpicommander")
"""Global application settings."""

homedir = environ.get("HOME", "") or environ.get("APPDATA", "")
"""Home dir of user."""


class ConnectionFail(IntEnum):
    NO_XML_RPC = 1
    SSH_CONNECT = 2
    SSH_AUTH = 4
    NO_XML_RPC_VIA_TUNNEL = 9  # Includes NO_XML_RPC Bit


class WidgetData(IntEnum):
    address = 260
    acl_level = 262
    has_error = 263
    port = 264
    object_name = 265
    host_name = 267
    host_name_full = 268
    file_name = 309
    revpi_settings = 320


class RevPiSettings:

    def __init__(self, load_index: int = None, settings_storage: QtCore.QSettings = None):
        """
        Revolution Pi saved settings.

        :param load_index: Load settings from index, same as .load_from_index
        :param settings_storage: Change QSettings object to work on from default to this one
        """
        self._settings = settings_storage or settings
        self.internal_id = ""

        self.name = "New connection"
        self.folder = ""
        self.address = "127.0.0.1"
        self.port = 55123
        self.timeout = 5

        self.ssh_use_tunnel = True
        self.ssh_port = 22
        self.ssh_user = "pi"
        self.ssh_saved_password = False

        self.last_dir_upload = "."
        self.last_file_upload = "."
        self.last_dir_pictory = "."
        self.last_dir_picontrol = "."
        self.last_dir_selected = "."
        self.last_pictory_file = ""
        self.last_tar_file = ""
        self.last_zip_file = ""
        self.watch_files = []
        self.watch_path = ""

        self.debug_geos = {}

        if load_index is not None:
            self.load_from_index(load_index)

    def load_from_index(self, settings_index: int) -> None:
        """Load settings from 'connections' index."""
        self._settings.beginReadArray("connections")
        self._settings.setArrayIndex(settings_index)

        # Flag as "legacy" connection to generate missing internal_id on save_settings()
        self.internal_id = self._settings.value("internal_id", "legacy", type=str)

        self.name = self._settings.value("name", type=str)
        self.folder = self._settings.value("folder", "", type=str)
        self.address = self._settings.value("address", type=str)
        self.port = self._settings.value("port", 55123, type=int)
        self.timeout = self._settings.value("timeout", 5, type=int)

        self.ssh_use_tunnel = self._settings.value("ssh_use_tunnel", True, type=bool)
        self.ssh_port = self._settings.value("ssh_port", 22, type=int)
        self.ssh_user = self._settings.value("ssh_user", "pi", type=str)
        self.ssh_saved_password = self._settings.value("ssh_saved_password", False, type=bool)

        self.last_dir_upload = self._settings.value("last_dir_upload", ".", type=str)
        self.last_file_upload = self._settings.value("last_file_upload", ".", type=str)
        self.last_dir_pictory = self._settings.value("last_dir_pictory", ".", type=str)
        self.last_dir_picontrol = self._settings.value("last_dir_picontrol", ".", type=str)
        self.last_dir_selected = self._settings.value("last_dir_selected", ".", type=str)
        self.last_pictory_file = self._settings.value("last_pictory_file", "", type=str)
        self.last_tar_file = self._settings.value("last_tar_file", "", type=str)
        self.last_zip_file = self._settings.value("last_zip_file", "", type=str)
        self.watch_files = self._settings.value("watch_files", [], type=list)
        self.watch_path = self._settings.value("watch_path", "", type=str)

        try:
            # Bytes with QSettings are a little difficult sometimes
            self.debug_geos = self._settings.value("debug_geos", {}, type=dict)
        except Exception:
            # Just drop the geos of IO windows
            pass

        # These values must exists
        if not (self.name and self.address and self.port):
            raise ValueError("Could not geht all required values from saved settings")

        self._settings.endArray()

    def save_settings(self):
        """Save all settings."""

        count_settings = self._settings.beginReadArray("connections")

        def create_new_array_member():
            """Insert a new setting at the end of the array."""

            # Close the active array action to reopen a write action to expand the array
            self._settings.endArray()
            self._settings.beginWriteArray("connections")
            self._settings.setArrayIndex(count_settings)

            if not self.internal_id:
                self.internal_id = uuid4().hex

        if not self.internal_id:
            create_new_array_member()

        else:
            # Always search setting in array, because connection manager could reorganize array indexes
            new_setting = True
            for index in range(count_settings):
                self._settings.setArrayIndex(index)

                if self.internal_id == "legacy":
                    # Legacy connection without internal_id
                    if self._settings.value("address") == self.address:
                        # Set missing internal_id
                        self.internal_id = uuid4().hex
                        new_setting = False
                        break
                else:
                    if self._settings.value("internal_id") == self.internal_id:
                        new_setting = False
                        break

            if new_setting:
                # On this point, we iterate all settings and found none, so create new one
                create_new_array_member()

        self._settings.setValue("internal_id", self.internal_id)
        self._settings.setValue("name", self.name)
        self._settings.setValue("folder", self.folder)
        self._settings.setValue("address", self.address)
        self._settings.setValue("port", self.port)
        self._settings.setValue("timeout", self.timeout)

        self._settings.setValue("ssh_use_tunnel", self.ssh_use_tunnel)
        self._settings.setValue("ssh_port", self.ssh_port)
        self._settings.setValue("ssh_user", self.ssh_user)
        self._settings.setValue("ssh_saved_password", self.ssh_saved_password)

        self._settings.setValue("last_dir_upload", self.last_dir_upload)
        self._settings.setValue("last_file_upload", self.last_file_upload)
        self._settings.setValue("last_dir_pictory", self.last_dir_pictory)
        self._settings.setValue("last_dir_picontrol", self.last_dir_picontrol)
        self._settings.setValue("last_dir_selected", self.last_dir_selected)
        self._settings.setValue("last_pictory_file", self.last_pictory_file)
        self._settings.setValue("last_tar_file", self.last_tar_file)
        self._settings.setValue("last_zip_file", self.last_zip_file)
        self._settings.setValue("watch_files", self.watch_files)
        self._settings.setValue("watch_path", self.watch_path)
        self._settings.setValue("debug_geos", self.debug_geos)

        self._settings.endArray()


class ConnectionManager(QtCore.QThread):
    """Check connection and status for PLC program on Revolution Pi."""

    connect_error = QtCore.pyqtSignal(str, str, ConnectionFail, RevPiSettings)
    """Error header, message and reason (ConnectionFail) of a new connection after pyload_connect call."""
    connection_established = QtCore.pyqtSignal()
    """New connection established successfully with <class 'ServerProxy'>."""
    connection_disconnected = QtCore.pyqtSignal()
    """Connection disconnected."""
    connection_disconnecting = QtCore.pyqtSignal()
    """Signal emitted before closing connection."""
    connection_error_observed = QtCore.pyqtSignal(str)
    """This will be triggered, if a connection error was detected."""
    status_changed = QtCore.pyqtSignal(str, str)
    """Status message and color suggestion."""
    connection_recovered = QtCore.pyqtSignal()
    """After errors the connection is established again, could have other port information (SSH)."""

    def __init__(self, parent=None, cycle_time_ms=1000):
        super().__init__(parent)

        self._cli = None
        self._cli_connect = Queue()
        self._cycle_time = cycle_time_ms
        self._lck_cli = Lock()
        self._ps_started = False
        self._revpi = None
        self._revpi_output = None

        self.settings = RevPiSettings()

        self.ssh_tunnel_server = None  # type: SSHLocalTunnel
        self.ssh_pass = ""

        self.pyload_version = (0, 0, 0)
        """Version number of RevPiPyLoad 0.0.0 with <class 'int'> values."""
        self.pyload_version_str = ""
        """Raw string of RevPyPyLoad version, could contain rc1 at the end."""
        self.xml_funcs = []
        """Name list of all supported functions of RevPiPyLoad."""
        self.xml_mode = -1
        """ACL level for this connection (-1 on connection denied)."""
        self._xml_mode_refresh = False

    def __call_simulator(self, function_name: str, *args, default_value=None, **kwargs):
        pi.logger.debug("ConnectionManager.__call_simulator({0})".format(function_name))
        if function_name == "ps_values":
            if self._revpi.readprocimg():
                bytebuff = bytearray()
                for dev in self._revpi.device:
                    bytebuff += bytes(dev)
                return Binary(bytes(bytebuff))

        elif function_name == "ps_setvalue":
            # args: 0=device, 1=io, 2=value
            device = args[0]
            io = args[1]
            if type(args[2]) == Binary:
                value = args[2].data
            else:
                value = args[2]

            try:
                # Write new value to IO
                self._revpi.io[io].set_value(value)
            except Exception as e:
                return [device, io, False, str(e)]

            return [device, io, True, ""]

        elif function_name == "psstart":
            self._revpi.autorefresh_all()
            return True

        elif function_name == "psstop":
            self._revpi.exit(full=False)
            return True

        elif function_name == "ps_devices":
            return [(dev.position, dev.name) for dev in self._revpi.device]

        elif function_name == "ps_inps":
            return self.__simulator_ios("inp")

        elif function_name == "ps_outs":
            return self.__simulator_ios("out")

        else:
            return default_value

    def __simulator_ios(self, iotype: str):
        dict_ios = {}
        for dev in self._revpi.device:
            dict_ios[dev.position] = []

            if iotype == "inp":
                lst_io = dev.get_inputs()
            elif iotype == "out":
                lst_io = dev.get_outputs()
            else:
                lst_io = []

            for io in lst_io:
                dict_ios[dev.position].append([
                    io.name,
                    1 if io._bitlength == 1 else int(io._bitlength / 8),
                    io._slc_address.start + dev.offset,
                    io.bmk,
                    io._bitaddress,
                    io._byteorder,
                    io._signed,
                ])
        return Binary(pickle.dumps(dict_ios))

    def _clear_settings(self):
        """Clear connection settings."""
        self.settings = RevPiSettings()

        self.ssh_pass = ""

        self.pyload_version = (0, 0, 0)
        self.pyload_version_str = ""
        self.xml_funcs.clear()
        self.xml_mode = -1

    def pyload_connect(self, revpi_settings: RevPiSettings, ssh_pass="") -> bool:
        """
        Create a new connection from settings object.

        :param revpi_settings: Revolution Pi saved connection settings
        :param ssh_pass: Use this ssh password, if revpi_settings.ssh_use_tunnel is true
        :return: True, if the connection was successfully established
        """

        # First disconnect to send signal and clean up values
        self.pyload_disconnect()

        ssh_tunnel_server = None
        ssh_tunnel_port = 0

        socket.setdefaulttimeout(revpi_settings.timeout)

        if revpi_settings.ssh_use_tunnel:
            ssh_tunnel_server = SSHLocalTunnel(
                revpi_settings.port,
                revpi_settings.address,
                revpi_settings.ssh_port
            )
            try:
                ssh_tunnel_port = ssh_tunnel_server.connect_by_credentials(revpi_settings.ssh_user, ssh_pass)

                if getattr(revpi_settings, "ssh_enable_revpipyload", False):
                    ssh_tunnel_server.send_cmd("sudo systemctl enable --now revpipyload")

            except AuthenticationException:
                self.connect_error.emit(
                    self.tr("Error"), self.tr(
                        "The combination of username and password was rejected from the SSH server.\n\n"
                        "Try again."
                    ),
                    ConnectionFail.SSH_AUTH,
                    revpi_settings,
                )
                return False
            except Exception as e:
                # todo: Check some more kinds of exceptions and nice user info
                self._clear_settings()
                self.connect_error.emit(
                    self.tr("Error"), self.tr(
                        "Could not establish a SSH connection to server:\n\n{0}"
                    ).format(str(e)),
                    ConnectionFail.SSH_CONNECT,
                    revpi_settings,
                )
                return False

            sp = ServerProxy("http://127.0.0.1:{0}".format(ssh_tunnel_port))

        else:
            sp = ServerProxy("http://{0}:{1}".format(revpi_settings.address, revpi_settings.port))

        # Load values and test connection to Revolution Pi
        try:
            ma = search(r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)", sp.version())
            pyload_version = int(ma.group("major")), int(ma.group("minor")), int(ma.group("patch"))
            pyload_version_str = ma.string
            xml_funcs = sp.system.listMethods()
            xml_mode = sp.xmlmodus()
        except Exception as e:
            pi.logger.exception(e)
            self.connection_error_observed.emit(str(e))

            if revpi_settings.ssh_use_tunnel:
                self.connect_error.emit(
                    self.tr("Error"), self.tr(
                        "Can not connect to RevPiPyLoad service through SSH tunnel!\n\n"
                        "This could have the following reasons:\n"
                        "- The RevPiPyLoad service is not running (activate it on your Revolution Pi)\n"
                        "- The RevPiPyLoad XML-RPC service is NOT bind to localhost\n"
                        "- The ACL permission is not set for 127.0.0.1!!!"
                    ),
                    ConnectionFail.NO_XML_RPC_VIA_TUNNEL,
                    revpi_settings,
                )
            else:
                self.connect_error.emit(
                    self.tr("Error"), self.tr(
                        "Can not connect to RevPiPyLoad XML-RPC service! \n\n"
                        "This could have the following reasons:\n"
                        "- The Revolution Pi is not online\n"
                        "- The RevPiPyLoad service is not running (activate it on your Revolution Pi)\n"
                        "- The RevPiPyLoad XML-RPC service is bind to localhost, only\n"
                        "- The ACL permission is not set for your IP!!!\n\n"
                        "Use 'Connect via SSH' to use an encrypted connection or run "
                        "'sudo revpipyload_secure_installation' on Revolution Pi to setup direct remote access!"
                    ),
                    ConnectionFail.NO_XML_RPC,
                    revpi_settings,
                )

            return False

        self.settings = revpi_settings
        self.ssh_pass = ssh_pass
        self.pyload_version = pyload_version
        self.pyload_version_str = pyload_version_str
        self.xml_funcs = xml_funcs
        self.xml_mode = xml_mode

        with self._lck_cli:
            self.ssh_tunnel_server = ssh_tunnel_server
            self._cli = sp
            self._cli_connect.put_nowait((
                "127.0.0.1" if revpi_settings.ssh_use_tunnel else revpi_settings.address,
                ssh_tunnel_port if revpi_settings.ssh_use_tunnel else revpi_settings.port
            ))

        self.connection_established.emit()

        return True

    def pyload_disconnect(self):
        """Disconnect from Revolution Pi."""
        if self._revpi is not None:
            self.connection_disconnecting.emit()

            self._revpi.cleanup()
            self._revpi_output.cleanup()
            if settings.value("simulator/stop_remove", False, bool):
                remove(self._revpi.procimg)
            self._revpi = None
            self._revpi_output = None

            pi.logger.debug("Simulator destroyed.")
            self.connection_disconnected.emit()

        elif self._cli is not None:

            # Tell all widget, that we want to disconnect
            self.connection_disconnecting.emit()
            self.settings.save_settings()

            with self._lck_cli:
                if self._ps_started:
                    try:
                        self._cli.psstop()
                    except Exception:
                        pass
                self._clear_settings()
                self._cli = None

                if self.ssh_tunnel_server:
                    self.ssh_tunnel_server.disconnect()
                    self.ssh_tunnel_server = None

            self.connection_disconnected.emit()

    def pyload_simulate(self, configrsc: str, procimg: str, clean_existing: bool):
        """
        Start the simulator for piControl on local computer.

        :param configrsc: piCtory configuration
        :param procimg: Process image, which is a 4 kByte file for simulation
        :param clean_existing: Reset the file to ZERO \x00 bytes
        """
        pi.logger.debug("ConnectionManager.start_simulate")

        if not exists(procimg) or clean_existing:
            with open(procimg, "wb") as fh:
                fh.write(b'\x00' * 4096)

        try:
            import revpimodio2

            # Prepare process image with default values for outputs
            self._revpi_output = revpimodio2.RevPiModIO(configrsc=configrsc, procimg=procimg)
            self._revpi_output.setdefaultvalues()
            self._revpi_output.writeprocimg()

            # This is our simulator to work with
            self._revpi = revpimodio2.RevPiModIO(simulator=True, configrsc=configrsc, procimg=procimg)
            self._revpi.setdefaultvalues()
            self._revpi.writeprocimg()

            self.xml_funcs = ["psstart", "psstop", "ps_devices", "ps_inps", "ps_outs", "ps_values", "ps_setvalue"]

            self.connection_established.emit()

        except Exception as e:
            pi.logger.exception(e)
            self.connection_error_observed.emit(str(e))
            self._revpi_output = None
            self._revpi = None
            if settings.value("simulator/stop_remove", False, bool):
                remove(procimg)

        return self._revpi is not None

    def refresh_xml_mode(self):
        """Refresh XML ACL level after some change could be done."""
        self._xml_mode_refresh = True

    def reset_simulator(self):
        """Reset all io to piCtory defaults."""
        pi.logger.debug("ConnectionManager.reset_simulator")
        if settings.value("simulator/restart_zero", False, bool):
            with open(self._revpi.procimg, "wb") as fh:
                fh.write(b'\x00' * 4096)
            self._revpi.readprocimg()
        else:
            self._revpi_output.writeprocimg()
            self._revpi.setdefaultvalues()
            self._revpi.writeprocimg()

    def run(self):
        """Thread worker to check status of RevPiPyLoad."""
        self.setPriority(QtCore.QThread.NormalPriority)

        sp = None
        while not self.isInterruptionRequested():

            if self._revpi is not None:
                sp = None
                self.status_changed.emit(self.tr("SIMULATING"), "yellow")
            elif self._cli is None:
                sp = None
                self.status_changed.emit(self.tr("NOT CONNECTED"), "lightblue")
            elif not self._cli_connect.empty():
                # Get new connection information to create object in this thread
                item = self._cli_connect.get()
                sp = ServerProxy("http://{0}:{1}".format(*item))
                self._cli_connect.task_done()

            if sp:
                try:
                    plc_exit_code = sp.plcexitcode()
                    if self._xml_mode_refresh:
                        self.xml_mode = sp.xmlmodus()
                        self._xml_mode_refresh = False
                except CannotSendRequest as e:
                    pi.logger.warning(e)
                except Exception as e:
                    pi.logger.warning(e)
                    self.status_changed.emit(self.tr("SERVER ERROR"), "red")
                    self.connection_error_observed.emit("{0} | {1}".format(e, type(e)))

                    if self.ssh_tunnel_server and not self.ssh_tunnel_server.connected:
                        self.ssh_tunnel_server.disconnect()
                        ssh_tunnel_server = SSHLocalTunnel(
                            self.settings.port,
                            self.settings.address,
                            self.settings.ssh_port
                        )
                        try:
                            ssh_tunnel_port = self.ssh_tunnel_server.connect_by_credentials(
                                self.settings.ssh_user,
                                self.ssh_pass
                            )
                            sp = ServerProxy("http://127.0.0.1:{0}".format(ssh_tunnel_port))
                            with self._lck_cli:
                                self.ssh_tunnel_server = ssh_tunnel_server
                                self._cli = sp
                                self.connection_recovered.emit()
                        except Exception:
                            pass

                else:
                    if plc_exit_code == -1:
                        self.status_changed.emit(self.tr("RUNNING"), "green")
                    elif plc_exit_code == -2:
                        self.status_changed.emit(self.tr("PLC FILE NOT FOUND"), "red")
                    elif plc_exit_code == -3:
                        self.status_changed.emit(self.tr("NOT RUNNING (NO STATUS)"), "yellow")
                    elif plc_exit_code == -9:
                        self.status_changed.emit(self.tr("PROGRAM KILLED"), "red")
                    elif plc_exit_code == -15:
                        self.status_changed.emit(self.tr("PROGRAM TERMED"), "red")
                    elif plc_exit_code == 0:
                        self.status_changed.emit(self.tr("NOT RUNNING"), "yellow")
                    else:
                        self.status_changed.emit(self.tr("FINISHED WITH CODE {0}").format(plc_exit_code), "yellow")

            self.msleep(self._cycle_time)

    def call_remote_function(self, function_name: str, *args, default_value=None, raise_exception=False, **kwargs):
        """
        Save call of a remote function with given name and parameters on Revolution Pi.

        :param function_name: Function to call on RevPiPyLoad
        :param args: Functions arguments
        :param default_value: Default value will be returned on error
        :param raise_exception: Will raise the exception returned from server
        :param kwargs: Functions key word arguments
        :return: Return value of remote function or default_value
        """
        if self._cli is None and self._revpi is None:
            pi.logger.error("Not connected while calling {0}".format(function_name))
            if raise_exception:
                raise ConnectionError("Connection manager not connected")
            return default_value

        reload_funcs = False
        if function_name == "psstart":
            self._ps_started = True
            reload_funcs = True
        elif function_name == "psstop":
            self._ps_started = False
            reload_funcs = True

        # On connection problems do not freeze
        if self._lck_cli.acquire(timeout=1.0):
            if self._revpi is not None:
                # Redirect call to simulator
                return_value = self.__call_simulator(function_name, *args, default_value=default_value, **kwargs)
            else:
                try:
                    return_value = getattr(self._cli, function_name)(*args, **kwargs)
                    if reload_funcs:
                        self.xml_funcs = self._cli.system.listMethods()
                except Exception as e:
                    pi.logger.error(e)
                    if raise_exception:
                        self._lck_cli.release()
                        raise
                    return_value = default_value

            self._lck_cli.release()
            return return_value

        elif raise_exception:
            raise ConnectionError("Can not get lock of connection")

        return default_value

    def get_cli(self):
        """
        Connection proxy of actual connection.

        Use connection_recovered signal to figure out new parameters.
        """
        if not self.settings.ssh_use_tunnel and self.settings.address and self.settings.port:
            return ServerProxy("http://{0}:{1}".format(self.settings.address, self.settings.port))
        if self.settings.ssh_use_tunnel and self.ssh_tunnel_server and self.ssh_tunnel_server.connected:
            return ServerProxy("http://127.0.0.1:{0}".format(self.ssh_tunnel_server.local_tunnel_port))

        return None

    @property
    def connected(self) -> bool:
        """True if we have an active connection."""
        return self._cli is not None

    @property
    def simulating(self) -> bool:
        """True, if simulating mode is running."""
        return self._revpi is not None

    @property
    def simulating_configrsc(self) -> str:
        return self._revpi.configrsc if self._revpi else ""

    @property
    def simulating_procimg(self) -> str:
        return self._revpi.procimg if self._revpi else ""


cm = ConnectionManager()
"""Clobal connection manager instance."""


def all_revpi_settings() -> [RevPiSettings]:
    """Get all revpi settings objects."""
    # Get length of array and close it, the RevPiSettings-class need it
    count_settings = settings.beginReadArray("connections")
    settings.endArray()
    return [RevPiSettings(i) for i in range(count_settings)]


def swap_word_order(bytes_to_swap) -> bytes:
    """Swap word order of an even byte array."""
    array_lenght = len(bytes_to_swap)
    swap_array = bytearray(bytes_to_swap)
    for i in range(0, array_lenght // 2, 2):
        swap_array[-i - 2:array_lenght - i], swap_array[i:i + 2] = \
            swap_array[i:i + 2], swap_array[-i - 2:array_lenght - i]
    return bytes(swap_array)


def import_old_settings():
    """Try to import saved connections from old storage to new setting object."""
    if settings.value("revpicommander/imported_settings", False, type=bool):
        return
    settings.setValue("revpicommander/imported_settings", True)

    old_settings = QtCore.QSettings("revpipyplc", "revpipyload")
    count_settings = old_settings.beginReadArray("connections")
    old_settings.endArray()

    for i in range(count_settings):
        try:
            revpi_setting = RevPiSettings(i, settings_storage=old_settings)
            revpi_setting._settings = settings
            revpi_setting.save_settings()
        except Exception:
            pi.logger.warning("Could not import saved connection {0}".format(i))


import_old_settings()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'simulator.ui'
#
# Created by: PyQt5 UI code generator 5.15.7
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_diag_simulator(object):
    def setupUi(self, diag_simulator):
        diag_simulator.setObjectName("diag_simulator")
        diag_simulator.resize(522, 500)
        self.verticalLayout = QtWidgets.QVBoxLayout(diag_simulator)
        self.verticalLayout.setObjectName("verticalLayout")
        self.gb_settings = QtWidgets.QGroupBox(diag_simulator)
        self.gb_settings.setObjectName("gb_settings")
        self.gridLayout = QtWidgets.QGridLayout(self.gb_settings)
        self.gridLayout.setObjectName("gridLayout")
        self.lbl_history = QtWidgets.QLabel(self.gb_settings)
        self.lbl_history.setObjectName("lbl_history")
        self.gridLayout.addWidget(self.lbl_history, 0, 0, 1, 1)
        self.lbl_configrsc = QtWidgets.QLabel(self.gb_settings)
        self.lbl_configrsc.setObjectName("lbl_configrsc")
        self.gridLayout.addWidget(self.lbl_configrsc, 1, 0, 1, 1)
        self.txt_configrsc = QtWidgets.QLineEdit(self.gb_settings)
        self.txt_configrsc.setFocusPolicy(QtCore.Qt.NoFocus)
        self.txt_configrsc.setText("")
        self.txt_configrsc.setObjectName("txt_configrsc")
        self.gridLayout.addWidget(self.txt_configrsc, 1, 1, 1, 1)
        self.btn_configrsc = QtWidgets.QPushButton(self.gb_settings)
        self.btn_configrsc.setObjectName("btn_configrsc")
        self.gridLayout.addWidget(self.btn_configrsc, 1, 2, 1, 1)
        self.lbl_procimg = QtWidgets.QLabel(self.gb_settings)
        self.lbl_procimg.setObjectName("lbl_procimg")
        self.gridLayout.addWidget(self.lbl_procimg, 2, 0, 1, 1)
        self.txt_procimg = QtWidgets.QLineEdit(self.gb_settings)
        self.txt_procimg.setFocusPolicy(QtCore.Qt.NoFocus)
        self.txt_procimg.setText("")
        self.txt_procimg.setObjectName("txt_procimg")
        self.gridLayout.addWidget(self.txt_procimg, 2, 1, 1, 1)
        self.lbl_stop = QtWidgets.QLabel(self.gb_settings)
        self.lbl_stop.setObjectName("lbl_stop")
        self.gridLayout.addWidget(self.lbl_stop, 3, 0, 1, 1)
        self.lbl_restart = QtWidgets.QLabel(self.gb_settings)
        self.lbl_restart.setObjectName("lbl_restart")
        self.gridLayout.addWidget(self.lbl_restart, 4, 0, 1, 1)
        self.cbb_history = QtWidgets.QComboBox(self.gb_settings)
        self.cbb_history.setSizeAdjustPolicy(QtWidgets.QComboBox.AdjustToMinimumContentsLength)
        self.cbb_history.setObjectName("cbb_history")
        self.gridLayout.addWidget(self.cbb_history, 0, 1, 1, 2)
        self.cbx_stop_remove = QtWidgets.QCheckBox(self.gb_settings)
        self.cbx_stop_remove.setObjectName("cbx_stop_remove")
        self.gridLayout.addWidget(self.cbx_stop_remove, 3, 1, 1, 2)
        self.rb_restart_pictory = QtWidgets.QRadioButton(self.gb_settings)
        self.rb_restart_pictory.setChecked(True)
        self.rb_restart_pictory.setObjectName("rb_restart_pictory")
        self.gridLayout.addWidget(self.rb_restart_pictory, 4, 1, 1, 2)
        self.rb_restart_zero = QtWidgets.QRadioButton(self.gb_settings)
        self.rb_restart_zero.setObjectName("rb_restart_zero")
        self.gridLayout.addWidget(self.rb_restart_zero, 5, 1, 1, 2)
        self.verticalLayout.addWidget(self.gb_settings)
        self.gb_info = QtWidgets.QGroupBox(diag_simulator)
        self.gb_info.setObjectName("gb_info")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.gb_info)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.lbl_info = QtWidgets.QLabel(self.gb_info)
        self.lbl_info.setWordWrap(True)
        self.lbl_info.setObjectName("lbl_info")
        self.verticalLayout_2.addWidget(self.lbl_info)
        self.txt_info = QtWidgets.QPlainTextEdit(self.gb_info)
        self.txt_info.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.AdjustToContents)
        self.txt_info.setReadOnly(True)
        self.txt_info.setPlainText("")
        self.txt_info.setObjectName("txt_info")
        self.verticalLayout_2.addWidget(self.txt_info)
        self.verticalLayout.addWidget(self.gb_info)
        self.btn_start_pictory = QtWidgets.QPushButton(diag_simulator)
        self.btn_start_pictory.setShortcut("Ctrl+1")
        self.btn_start_pictory.setObjectName("btn_start_pictory")
        self.verticalLayout.addWidget(self.btn_start_pictory)
        self.btn_start_empty = QtWidgets.QPushButton(diag_simulator)
        self.btn_start_empty.setShortcut("Ctrl+2")
        self.btn_start_empty.setObjectName("btn_start_empty")
        self.verticalLayout.addWidget(self.btn_start_empty)
        self.btn_start_nochange = QtWidgets.QPushButton(diag_simulator)
        self.btn_start_nochange.setShortcut("Ctrl+3")
        self.btn_start_nochange.setObjectName("btn_start_nochange")
        self.verticalLayout.addWidget(self.btn_start_nochange)

        self.retranslateUi(diag_simulator)
        self.btn_start_empty.clicked.connect(diag_simulator.accept) # type: ignore
        self.btn_start_nochange.clicked.connect(diag_simulator.accept) # type: ignore
        self.btn_start_pictory.clicked.connect(diag_simulator.accept) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(diag_simulator)
        diag_simulator.setTabOrder(self.cbb_history, self.btn_configrsc)
        diag_simulator.setTabOrder(self.btn_configrsc, self.cbx_stop_remove)
        diag_simulator.setTabOrder(self.cbx_stop_remove, self.rb_restart_pictory)
        diag_simulator.setTabOrder(self.rb_restart_pictory, self.rb_restart_zero)
        diag_simulator.setTabOrder(self.rb_restart_zero, self.txt_info)
        diag_simulator.setTabOrder(self.txt_info, self.btn_start_pictory)
        diag_simulator.setTabOrder(self.btn_start_pictory, self.btn_start_empty)
        diag_simulator.setTabOrder(self.btn_start_empty, self.btn_start_nochange)

    def retranslateUi(self, diag_simulator):
        _translate = QtCore.QCoreApplication.translate
        diag_simulator.setWindowTitle(_translate("diag_simulator", "piControl simulator"))
        self.gb_settings.setTitle(_translate("diag_simulator", "Simulator settings"))
        self.lbl_history.setText(_translate("diag_simulator", "Last used:"))
        self.lbl_configrsc.setText(_translate("diag_simulator", "piCtory file:"))
        self.btn_configrsc.setText(_translate("diag_simulator", "select..."))
        self.lbl_procimg.setText(_translate("diag_simulator", "Process image:"))
        self.lbl_stop.setText(_translate("diag_simulator", "Stop action:"))
        self.lbl_restart.setText(_translate("diag_simulator", "Restart action:"))
        self.cbx_stop_remove.setText(_translate("diag_simulator", "Remove process image file"))
        self.rb_restart_pictory.setText(_translate("diag_simulator", "Restore piCtory default values"))
        self.rb_restart_zero.setText(_translate("diag_simulator", "Reset everything to ZERO"))
        self.gb_info.setTitle(_translate("diag_simulator", "RevPiModIO integration"))
        self.lbl_info.setText(_translate("diag_simulator", "You can work with this simulator if you call RevPiModIO with this additional parameters:"))
        self.btn_start_pictory.setText(_translate("diag_simulator", "Start with piCtory default values"))
        self.btn_start_empty.setText(_translate("diag_simulator", "Start with empty process image"))
        self.btn_start_nochange.setText(_translate("diag_simulator", "Start without changing actual process image"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    diag_simulator = QtWidgets.QDialog()
    ui = Ui_diag_simulator()
    ui.setupUi(diag_simulator)
    diag_simulator.show()
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'debugios.ui'
#
# Created by: PyQt5 UI code generator 5.15.7
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_win_debugios(object):
    def setupUi(self, win_debugios):
        win_debugios.setObjectName("win_debugios")
        win_debugios.resize(434, 343)
        self.centralwidget = QtWidgets.QWidget(win_debugios)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout.setObjectName("verticalLayout")
        self.gb_io = QtWidgets.QGroupBox(self.centralwidget)
        self.gb_io.setObjectName("gb_io")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.gb_io)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.splitter = QtWidgets.QSplitter(self.gb_io)
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.splitter.setChildrenCollapsible(False)
        self.splitter.setObjectName("splitter")
        self.sca_inp = QtWidgets.QScrollArea(self.splitter)
        self.sca_inp.setLineWidth(0)
        self.sca_inp.setWidgetResizable(True)
        self.sca_inp.setObjectName("sca_inp")
        self.saw_inp = QtWidgets.QWidget()
        self.saw_inp.setGeometry(QtCore.QRect(0, 0, 201, 275))
        self.saw_inp.setObjectName("saw_inp")
        self.form_inp = QtWidgets.QFormLayout(self.saw_inp)
        self.form_inp.setLabelAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.form_inp.setContentsMargins(-1, 6, -1, 6)
        self.form_inp.setObjectName("form_inp")
        self.sca_inp.setWidget(self.saw_inp)
        self.sca_out = QtWidgets.QScrollArea(self.splitter)
        self.sca_out.setLineWidth(0)
        self.sca_out.setWidgetResizable(True)
        self.sca_out.setObjectName("sca_out")
        self.saw_out = QtWidgets.QWidget()
        self.saw_out.setGeometry(QtCore.QRect(0, 0, 201, 275))
        self.saw_out.setObjectName("saw_out")
        self.form_out = QtWidgets.QFormLayout(self.saw_out)
        self.form_out.setLabelAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.form_out.setContentsMargins(-1, 6, -1, 6)
        self.form_out.setObjectName("form_out")
        self.sca_out.setWidget(self.saw_out)
        self.verticalLayout_2.addWidget(self.splitter)
        self.verticalLayout.addWidget(self.gb_io)
        win_debugios.setCentralWidget(self.centralwidget)
        self.stat_bar = QtWidgets.QStatusBar(win_debugios)
        self.stat_bar.setObjectName("stat_bar")
        win_debugios.setStatusBar(self.stat_bar)

        self.retranslateUi(win_debugios)
        QtCore.QMetaObject.connectSlotsByName(win_debugios)

    def retranslateUi(self, win_debugios):
        _translate = QtCore.QCoreApplication.translate
        self.gb_io.setTitle(_translate("win_debugios", "{0}: Inputs | Outputs"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    win_debugios = QtWidgets.QMainWindow()
    ui = Ui_win_debugios()
    ui.setupUi(win_debugios)
    win_debugios.show()
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-
"""Start main application of this package."""
__author__ = "Sven Sager"
__copyright__ = "Copyright (C) 2023 Sven Sager"
__license__ = "GPLv2"

# If we are running from a wheel, add the wheel to sys.path
if __package__ == "":
    from os.path import dirname
    from sys import path

    # __file__ is package-*.whl/package/__main__.py
    # Resulting path is the name of the wheel itself
    package_path = dirname(dirname(__file__))
    path.insert(0, package_path)

if __name__ == "__main__":
    import sys

    if len(sys.argv) == 2 and "--version" in sys.argv:
        # Catch --version, if this is the only argument (sys.argv[0] is always the script name)
        from revpicommander import __version__
        print(__version__)
        sys.exit(0)

    else:
        from revpicommander.revpicommander import main

        # Run the main application of this package
        sys.exit(main())



# -*- coding: utf-8 -*-
"""Manager for ACL lists."""
__author__ = "Sven Sager"
__copyright__ = "Copyright (C) 2023 Sven Sager"
__license__ = "GPLv2"

from re import compile

from PyQt5 import QtCore, QtGui, QtWidgets

from .helper import WidgetData
from .ui.aclmanager_ui import Ui_diag_aclmanager


class AclManager(QtWidgets.QDialog, Ui_diag_aclmanager):
    """ACL manager."""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setupUi(self)
        self.setFixedSize(self.size())

        self.__re_ipacl = compile(r"(25[0-5]|(2[0-4]|[01]?\d|)\d)(\.(25[0-5]|(2[0-4]|[01]?\d|)\d)){3},-1")
        self.__dict_acltext = {}
        self.__cbb_level_loaded_index = 0
        self.__mrk_message_shown = 0
        self.__oldacl = ""
        self.__read_only = False

        # Prepare GUI
        self.tb_acls.setColumnWidth(0, 300)
        self.tb_acls.setColumnWidth(1, 50)
        self.btn_edit.setEnabled(False)
        self.btn_remove.setEnabled(False)
        self.btn_add.setEnabled(False)

        # Move to next focus when enter a "."
        self.mrk_txt_ip_a_keyPressEvent = self.txt_ip_a.keyPressEvent
        self.mrk_txt_ip_b_keyPressEvent = self.txt_ip_b.keyPressEvent
        self.mrk_txt_ip_c_keyPressEvent = self.txt_ip_c.keyPressEvent
        self.mrk_txt_ip_d_keyPressEvent = self.txt_ip_d.keyPressEvent
        self.txt_ip_a.keyPressEvent = self.txt_ip_a_keyPressEvent
        self.txt_ip_b.keyPressEvent = self.txt_ip_b_keyPressEvent
        self.txt_ip_c.keyPressEvent = self.txt_ip_c_keyPressEvent
        self.txt_ip_d.keyPressEvent = self.txt_ip_d_keyPressEvent

    def __check_load_error(self):
        """
        Check load errors and shows a message one time.
        :return: True, if message was shown
        """
        if bool(self.__mrk_message_shown & 1):
            return False

        for row in range(self.tb_acls.rowCount()):
            if self.tb_acls.item(row, 0).data(WidgetData.has_error):
                self.__mrk_message_shown += 1
                QtWidgets.QMessageBox.critical(
                    self, self.tr("Error"), self.tr(
                        "There are errors in the ACL list!\nCheck the ALC levels of the "
                        "red lines in the table. The ACL levels or ip addresses are "
                        "invalid. If you save this dialog again, we will remove the "
                        "wrong entries automatically."
                    )
                )
                return True
        return False

    def _changes_done(self):
        """
        Check for unsaved changes in dialog.

        :return: True, if unsaved changes was found
        """
        return self.__table_to_acl() != self.__oldacl

    def accept(self) -> None:
        """Save settings."""
        if self.btn_add.isEnabled():
            # Entry is ready to save, did the user forgot to click the button?
            ask = QtWidgets.QMessageBox.question(
                self, self.tr("Unsaved entry"), self.tr(
                    "You worked on a new ACL entry. Do you want to save "
                    "that entry, too?"
                )
            ) == QtWidgets.QMessageBox.Yes
            if ask:
                self.on_btn_add_clicked()

        if self.__check_load_error():
            return

        self.__oldacl = self.__table_to_acl()
        super().accept()

    def closeEvent(self, a0: QtGui.QCloseEvent) -> None:
        if self._changes_done():
            ask = QtWidgets.QMessageBox.question(
                self, self.tr("Question"), self.tr(
                    "Do you really want to quit? \nUnsaved changes will be lost"
                )
            ) == QtWidgets.QMessageBox.Yes

            if ask:
                self.reject()
            else:
                a0.ignore()

    def exec(self) -> int:
        return super().exec()

    def reject(self) -> None:
        """Restore old settings."""
        self.setup_acl_manager(self.__oldacl, self.__dict_acltext)
        super().reject()

    def setup_acl_manager(self, acl_string: str, acl_texts: dict):
        if type(acl_string) != str:
            raise TypeError("acl_string must be <class 'str'>")
        if type(acl_texts) != dict:
            raise TypeError("acl_texts must be <class 'dict'>")

        self.__dict_acltext = acl_texts.copy()

        # Clean up widgets
        while self.tb_acls.rowCount() > 0:
            self.tb_acls.removeRow(0)
        self.cbb_level.clear()
        self.cbb_level.addItem(self.tr("Select..."), -1)
        self.lbl_level_info.clear()

        self.__re_ipacl = compile(
            r"([\d*]{1,3}\.){3}[\d*]{1,3},[" +
            str(min(self.__dict_acltext.keys(), default=0)) + r"-" +
            str(max(self.__dict_acltext.keys(), default=0)) + r"]"
        )

        for ip_level in acl_string.split(" "):
            self.__table_add_acl(ip_level)

        lst_level_text = []
        for level in sorted(self.__dict_acltext.keys()):
            level_text = self.tr("Level") + " {0}: {1}".format(level, self.__dict_acltext[level])
            lst_level_text.append(level_text)
            self.cbb_level.addItem(level_text, level)

        self.lbl_level_info.setText("\n".join(lst_level_text))
        self.__oldacl = self.__table_to_acl()

    def get_acl(self):
        """Get current ACL string."""
        return self.__oldacl

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    # region #      REGION: ACL list

    def __load_selected_entry(self):
        row = self.tb_acls.currentRow()
        if row < 0:
            return

        item = self.tb_acls.item(row, 0)
        ip_blocks = item.text().split(".")
        self.txt_ip_a.setText(ip_blocks[0])
        self.txt_ip_b.setText(ip_blocks[1])
        self.txt_ip_c.setText(ip_blocks[2])
        self.txt_ip_d.setText(ip_blocks[3])

        for index in range(self.cbb_level.count()):
            if self.cbb_level.itemData(index, QtCore.Qt.UserRole) == item.data(WidgetData.acl_level):
                self.__cbb_level_loaded_index = index
                self.cbb_level.setCurrentIndex(index)
                break

        self._check_all_filled()

    def __table_add_acl(self, ip_level: str):
        """Add ACL entry to table."""
        # Empty acl_string create empty ip_level
        if not ip_level:
            return

        ip, level = ip_level.split(",")
        if self.__re_ipacl.match(ip_level):
            brush = QtGui.QBrush()
            has_error = False
            tool_tip = ""
        else:
            brush = QtGui.QBrush(QtGui.QColor("red"))
            has_error = True
            tool_tip = self.tr("This entry has an invalid ACL level or wrong IP format!")

        for row in range(self.tb_acls.rowCount()):
            item_0 = self.tb_acls.item(row, 0)
            if item_0.text() == ip:
                item_1 = self.tb_acls.item(row, 1)

                # Update existing entry
                item_0.setData(WidgetData.acl_level, int(level))
                item_0.setData(WidgetData.has_error, has_error)
                item_0.setBackground(brush)
                item_0.setToolTip(tool_tip)
                item_1.setText(level)
                item_1.setBackground(brush)
                item_1.setToolTip(tool_tip)
                return

        row_count = self.tb_acls.rowCount()
        self.tb_acls.insertRow(row_count)

        item = QtWidgets.QTableWidgetItem(ip)
        item.setData(WidgetData.acl_level, int(level))
        item.setData(WidgetData.has_error, has_error)
        item.setBackground(brush)
        item.setToolTip(tool_tip)
        self.tb_acls.setItem(row_count, 0, item)

        item = QtWidgets.QTableWidgetItem(level)
        item.setBackground(brush)
        item.setToolTip(tool_tip)
        self.tb_acls.setItem(row_count, 1, item)

    def __table_to_acl(self, force=False, row_indexes=None):
        """
        Create acl string with valid entries only.

        :param force: If True, return all entries
        :param row_indexes: Only from indexes ist <class 'list'>
        :return: ACL string
        """
        if row_indexes is None:
            row_indexes = range(self.tb_acls.rowCount())

        buff_acl = ""
        for i in row_indexes:
            item = self.tb_acls.item(i, 0)
            ip_level = "{0},{1} ".format(item.text(), item.data(WidgetData.acl_level))
            if not (force or self.__re_ipacl.match(ip_level)):
                continue
            buff_acl += ip_level
        return buff_acl.strip()

    @QtCore.pyqtSlot(QtWidgets.QTableWidgetItem)
    def on_tb_acls_itemDoubleClicked(self, item: QtWidgets.QTableWidgetItem):
        if not self.__read_only:
            self.__load_selected_entry()

    @QtCore.pyqtSlot()
    def on_tb_acls_itemSelectionChanged(self):
        selected_rows = int(len(self.tb_acls.selectedItems()) / self.tb_acls.columnCount())
        self.btn_edit.setEnabled(not self.__read_only and selected_rows == 1)
        self.btn_remove.setEnabled(not self.__read_only and selected_rows > 0)

    @QtCore.pyqtSlot()
    def on_btn_edit_clicked(self):
        self.__load_selected_entry()

    @QtCore.pyqtSlot()
    def on_btn_remove_clicked(self):
        lst_selected_row_indexes = [mi.row() for mi in self.tb_acls.selectionModel().selectedRows(0)]
        if len(lst_selected_row_indexes) == 0:
            return

        str_remove = ""
        for ip_level in self.__table_to_acl(True, lst_selected_row_indexes).split(" "):
            if not ip_level:
                # Empty string will return empty field
                continue
            ip, level = ip_level.split(",")
            str_remove += "\nIP: {0:>15}\tLevel: {1}".format(ip, level)

        ask = QtWidgets.QMessageBox.question(
            self, self.tr("Question"), self.tr(
                "Do you really want to delete the following items?\n{0}"
            ).format(str_remove)
        ) == QtWidgets.QMessageBox.Yes

        if ask:
            # Turn order to start deleting from the button to preserve right indexes
            lst_selected_row_indexes.sort(reverse=True)
            for row in lst_selected_row_indexes:
                self.tb_acls.removeRow(row)

    # endregion # # # # #

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    # region #      REGION: Group box IP

    def _check_all_filled(self):
        all_filled = \
            len(self.txt_ip_a.text()) > 0 and \
            len(self.txt_ip_b.text()) > 0 and \
            len(self.txt_ip_c.text()) > 0 and \
            len(self.txt_ip_d.text()) > 0 and \
            self.cbb_level.currentIndex() != 0 and (
                self.txt_ip_a.isModified() or self.txt_ip_b.isModified() or
                self.txt_ip_c.isModified() or self.txt_ip_d.isModified() or
                self.cbb_level.currentIndex() != self.__cbb_level_loaded_index
            )

        self.btn_add.setEnabled(not self.__read_only and all_filled)

    def _move_ip_cursor(self, key: int, sender_widget: QtWidgets.QLineEdit,
                        last_widget: QtWidgets.QLineEdit, next_widget: QtWidgets.QLineEdit):
        """
        Move cursor between ip enter widgets.

        :param key: Pressed key to check
        :param sender_widget: Sender widget of this key
        :param last_widget: Set focus to this widget on backspace key
        :param next_widget: Set focus to this widget on period key
        :return: True, if the key should not be processed further
        """
        if last_widget and key == QtCore.Qt.Key_Backspace and len(sender_widget.text()) == 0:
            last_widget.setFocus()
        elif next_widget and key == QtCore.Qt.Key_Period:
            next_widget.setFocus()
            return True
        return False

    def txt_ip_a_keyPressEvent(self, a0: QtGui.QKeyEvent) -> None:
        if not self._move_ip_cursor(a0.key(), self.txt_ip_a, self.txt_ip_a, self.txt_ip_b):
            self.mrk_txt_ip_a_keyPressEvent(a0)

    def txt_ip_b_keyPressEvent(self, a0: QtGui.QKeyEvent) -> None:
        if not self._move_ip_cursor(a0.key(), self.txt_ip_b, self.txt_ip_a, self.txt_ip_c):
            self.mrk_txt_ip_b_keyPressEvent(a0)

    def txt_ip_c_keyPressEvent(self, a0: QtGui.QKeyEvent) -> None:
        if not self._move_ip_cursor(a0.key(), self.txt_ip_c, self.txt_ip_b, self.txt_ip_d):
            self.mrk_txt_ip_c_keyPressEvent(a0)

    def txt_ip_d_keyPressEvent(self, a0: QtGui.QKeyEvent) -> None:
        if not self._move_ip_cursor(a0.key(), self.txt_ip_d, self.txt_ip_c, self.txt_ip_d):
            self.mrk_txt_ip_d_keyPressEvent(a0)

    @QtCore.pyqtSlot(str)
    def on_txt_ip_a_textChanged(self, text: str):
        self._check_all_filled()

    @QtCore.pyqtSlot(str)
    def on_txt_ip_b_textChanged(self, text: str):
        self._check_all_filled()

    @QtCore.pyqtSlot(str)
    def on_txt_ip_c_textChanged(self, text: str):
        self._check_all_filled()

    @QtCore.pyqtSlot(str)
    def on_txt_ip_d_textChanged(self, text: str):
        self._check_all_filled()

    @QtCore.pyqtSlot(int)
    def on_cbb_level_currentIndexChanged(self, index: int):
        self._check_all_filled()

    @QtCore.pyqtSlot()
    def on_btn_add_clicked(self):
        """Add a new entry to acl table."""
        ip_level = "{0}.{1}.{2}.{3},{4}".format(
            self.txt_ip_a.text(),
            self.txt_ip_b.text(),
            self.txt_ip_c.text(),
            self.txt_ip_d.text(),
            self.cbb_level.currentData(QtCore.Qt.UserRole)
        )
        if self.__re_ipacl.match(ip_level):
            self.__table_add_acl(ip_level)
            self.on_btn_clear_clicked()
        else:
            QtWidgets.QMessageBox.critical(
                self, self.tr("Error"), self.tr(
                    "Can not save new ACL entry! Check format of ip address "
                    "and acl level is in value list."
                )
            )

    @QtCore.pyqtSlot()
    def on_btn_clear_clicked(self):
        """Clear entry widgets."""
        self.txt_ip_a.clear()
        self.txt_ip_b.clear()
        self.txt_ip_c.clear()
        self.txt_ip_d.clear()
        self.cbb_level.setCurrentIndex(0)
        self.__cbb_level_loaded_index = 0

    # endregion # # # # #

    @property
    def read_only(self):
        """Getter for read_only value."""
        return self.__read_only

    @read_only.setter
    def read_only(self, value):
        """Setter for read_only window."""
        self.__read_only = value
        self.txt_ip_a.setEnabled(not value)
        self.txt_ip_b.setEnabled(not value)
        self.txt_ip_c.setEnabled(not value)
        self.txt_ip_d.setEnabled(not value)
        self.cbb_level.setEnabled(not value)
        self.btn_clear.setEnabled(not value)
        if value:
            self.btn_box.setStandardButtons(
                QtWidgets.QDialogButtonBox.Cancel
            )
        else:
            self.btn_box.setStandardButtons(
                QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel
            )


# Debugging
if __name__ == '__main__':
    app = QtWidgets.QApplication([])
    win = AclManager(None)
    win.setup_acl_manager(
        "127.0.1.*,0 127.0.0.1,1 127.0.0.2,2", {
            0: "Just have a look",
            1: "Do more things",
        }
    )
    win.read_only = False
    rc = win.exec()
    print(
        "return code:", rc,
        "acl:", win.get_acl()
    )



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'mqttmanager.ui'
#
# Created by: PyQt5 UI code generator 5.15.7
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_diag_mqtt(object):
    def setupUi(self, diag_mqtt):
        diag_mqtt.setObjectName("diag_mqtt")
        diag_mqtt.resize(489, 709)
        self.verticalLayout = QtWidgets.QVBoxLayout(diag_mqtt)
        self.verticalLayout.setObjectName("verticalLayout")
        self.gb_basetopic = QtWidgets.QGroupBox(diag_mqtt)
        self.gb_basetopic.setObjectName("gb_basetopic")
        self.gridLayout = QtWidgets.QGridLayout(self.gb_basetopic)
        self.gridLayout.setObjectName("gridLayout")
        self.lbl_basetopic_description = QtWidgets.QLabel(self.gb_basetopic)
        self.lbl_basetopic_description.setMaximumSize(QtCore.QSize(450, 16777215))
        self.lbl_basetopic_description.setWordWrap(True)
        self.lbl_basetopic_description.setObjectName("lbl_basetopic_description")
        self.gridLayout.addWidget(self.lbl_basetopic_description, 1, 0, 1, 4)
        self.lbl_basetopic = QtWidgets.QLabel(self.gb_basetopic)
        self.lbl_basetopic.setObjectName("lbl_basetopic")
        self.gridLayout.addWidget(self.lbl_basetopic, 0, 0, 1, 1)
        self.txt_basetopic = QtWidgets.QLineEdit(self.gb_basetopic)
        self.txt_basetopic.setObjectName("txt_basetopic")
        self.gridLayout.addWidget(self.txt_basetopic, 0, 1, 1, 1)
        self.verticalLayout.addWidget(self.gb_basetopic)
        self.gb_send_on_event = QtWidgets.QGroupBox(diag_mqtt)
        self.gb_send_on_event.setObjectName("gb_send_on_event")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.gb_send_on_event)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.lbl_sendinterval = QtWidgets.QLabel(self.gb_send_on_event)
        self.lbl_sendinterval.setObjectName("lbl_sendinterval")
        self.gridLayout_2.addWidget(self.lbl_sendinterval, 0, 0, 1, 1)
        self.sbx_sendinterval = QtWidgets.QSpinBox(self.gb_send_on_event)
        self.sbx_sendinterval.setMinimum(1)
        self.sbx_sendinterval.setMaximum(21600)
        self.sbx_sendinterval.setObjectName("sbx_sendinterval")
        self.gridLayout_2.addWidget(self.sbx_sendinterval, 0, 1, 1, 1)
        self.lbl_topic_io = QtWidgets.QLabel(self.gb_send_on_event)
        self.lbl_topic_io.setObjectName("lbl_topic_io")
        self.gridLayout_2.addWidget(self.lbl_topic_io, 1, 0, 1, 2)
        self.cbx_send_on_event = QtWidgets.QCheckBox(self.gb_send_on_event)
        self.cbx_send_on_event.setObjectName("cbx_send_on_event")
        self.gridLayout_2.addWidget(self.cbx_send_on_event, 2, 0, 1, 2)
        self.lbl_topic_event = QtWidgets.QLabel(self.gb_send_on_event)
        self.lbl_topic_event.setObjectName("lbl_topic_event")
        self.gridLayout_2.addWidget(self.lbl_topic_event, 3, 0, 1, 2)
        self.verticalLayout.addWidget(self.gb_send_on_event)
        self.gb_write_outputs = QtWidgets.QGroupBox(diag_mqtt)
        self.gb_write_outputs.setObjectName("gb_write_outputs")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.gb_write_outputs)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.lbl_write_outputs = QtWidgets.QLabel(self.gb_write_outputs)
        self.lbl_write_outputs.setMaximumSize(QtCore.QSize(450, 16777215))
        self.lbl_write_outputs.setWordWrap(True)
        self.lbl_write_outputs.setObjectName("lbl_write_outputs")
        self.gridLayout_3.addWidget(self.lbl_write_outputs, 1, 0, 1, 1)
        self.cbx_write_outputs = QtWidgets.QCheckBox(self.gb_write_outputs)
        self.cbx_write_outputs.setObjectName("cbx_write_outputs")
        self.gridLayout_3.addWidget(self.cbx_write_outputs, 0, 0, 1, 1)
        self.verticalLayout.addWidget(self.gb_write_outputs)
        self.gb_broker = QtWidgets.QGroupBox(diag_mqtt)
        self.gb_broker.setObjectName("gb_broker")
        self.formLayout = QtWidgets.QFormLayout(self.gb_broker)
        self.formLayout.setObjectName("formLayout")
        self.lbl_broker_address = QtWidgets.QLabel(self.gb_broker)
        self.lbl_broker_address.setObjectName("lbl_broker_address")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.lbl_broker_address)
        self.txt_broker_address = QtWidgets.QLineEdit(self.gb_broker)
        self.txt_broker_address.setObjectName("txt_broker_address")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.txt_broker_address)
        self.lbl_port = QtWidgets.QLabel(self.gb_broker)
        self.lbl_port.setObjectName("lbl_port")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.lbl_port)
        self.lbl_username = QtWidgets.QLabel(self.gb_broker)
        self.lbl_username.setObjectName("lbl_username")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.lbl_username)
        self.lbl_password = QtWidgets.QLabel(self.gb_broker)
        self.lbl_password.setObjectName("lbl_password")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.lbl_password)
        self.lbl_client_id = QtWidgets.QLabel(self.gb_broker)
        self.lbl_client_id.setObjectName("lbl_client_id")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.LabelRole, self.lbl_client_id)
        self.txt_username = QtWidgets.QLineEdit(self.gb_broker)
        self.txt_username.setObjectName("txt_username")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.txt_username)
        self.txt_password = QtWidgets.QLineEdit(self.gb_broker)
        self.txt_password.setObjectName("txt_password")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.txt_password)
        self.txt_client_id = QtWidgets.QLineEdit(self.gb_broker)
        self.txt_client_id.setObjectName("txt_client_id")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.FieldRole, self.txt_client_id)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.sbx_port = QtWidgets.QSpinBox(self.gb_broker)
        self.sbx_port.setMinimum(1)
        self.sbx_port.setMaximum(65535)
        self.sbx_port.setObjectName("sbx_port")
        self.horizontalLayout.addWidget(self.sbx_port)
        self.cbx_tls_set = QtWidgets.QCheckBox(self.gb_broker)
        self.cbx_tls_set.setObjectName("cbx_tls_set")
        self.horizontalLayout.addWidget(self.cbx_tls_set)
        self.formLayout.setLayout(1, QtWidgets.QFormLayout.FieldRole, self.horizontalLayout)
        self.verticalLayout.addWidget(self.gb_broker)
        self.btn_box = QtWidgets.QDialogButtonBox(diag_mqtt)
        self.btn_box.setOrientation(QtCore.Qt.Horizontal)
        self.btn_box.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.btn_box.setObjectName("btn_box")
        self.verticalLayout.addWidget(self.btn_box)

        self.retranslateUi(diag_mqtt)
        self.btn_box.accepted.connect(diag_mqtt.accept) # type: ignore
        self.btn_box.rejected.connect(diag_mqtt.reject) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(diag_mqtt)

    def retranslateUi(self, diag_mqtt):
        _translate = QtCore.QCoreApplication.translate
        diag_mqtt.setWindowTitle(_translate("diag_mqtt", "MQTT settings"))
        self.gb_basetopic.setTitle(_translate("diag_mqtt", "Base topic"))
        self.lbl_basetopic_description.setText(_translate("diag_mqtt", "The base topic is the first part of any mqtt topic, the Revolution Pi will publish. You can use any character includig \'/\' to structure the messages on your broker.\n"
"\n"
"For example: revpi0000/data"))
        self.lbl_basetopic.setText(_translate("diag_mqtt", "Base topic:"))
        self.gb_send_on_event.setTitle(_translate("diag_mqtt", "Publish settings"))
        self.lbl_sendinterval.setText(_translate("diag_mqtt", "Publish all exported values every n seconds:"))
        self.lbl_topic_io.setText(_translate("diag_mqtt", "Topic:    [basetopic]/io/[ioname]"))
        self.cbx_send_on_event.setText(_translate("diag_mqtt", "Send exported values immediately on value change"))
        self.lbl_topic_event.setText(_translate("diag_mqtt", "Topic:    [basetopic]/event/[ioname]"))
        self.gb_write_outputs.setTitle(_translate("diag_mqtt", "Set outputs"))
        self.lbl_write_outputs.setText(_translate("diag_mqtt", "The Revolution Pi will subscribe a topic on which your mqtt client can publish messages with the new io value as payload.\n"
"\n"
"Publish values with topic:    [basetopic]/set/[outputname]"))
        self.cbx_write_outputs.setText(_translate("diag_mqtt", "Allow MQTT to to set outputs on Revolution Pi"))
        self.gb_broker.setTitle(_translate("diag_mqtt", "Broker settings"))
        self.lbl_broker_address.setText(_translate("diag_mqtt", "Broker address:"))
        self.lbl_port.setText(_translate("diag_mqtt", "Broker port:"))
        self.lbl_username.setText(_translate("diag_mqtt", "User name:"))
        self.lbl_password.setText(_translate("diag_mqtt", "Password:"))
        self.lbl_client_id.setText(_translate("diag_mqtt", "Client ID:"))
        self.cbx_tls_set.setText(_translate("diag_mqtt", "Use TLS"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    diag_mqtt = QtWidgets.QDialog()
    ui = Ui_diag_mqtt()
    ui.setupUi(diag_mqtt)
    diag_mqtt.show()
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-
"""Debug control widget to append to main window."""
__author__ = "Sven Sager"
__copyright__ = "Copyright (C) 2023 Sven Sager"
__license__ = "GPLv2"

import pickle
from xmlrpc.client import Binary, Fault, MultiCall, MultiCallIterator

from PyQt5 import QtCore, QtGui, QtWidgets

from . import helper
from . import proginit as pi
from .debugios import DebugIos
from .ui.debugcontrol_ui import Ui_wid_debugcontrol


class PsValues(QtCore.QThread):
    """
    Get process image from Revolution Pi.

    If this thread detects a driver reset, it will finish the work.
    """

    driver_reset_detected = QtCore.pyqtSignal()
    process_image_received = QtCore.pyqtSignal(Binary)

    def __init__(self):
        super().__init__()
        self._cycle_time = 200

    def run(self):
        """Read IO values of Revolution Pi."""
        pi.logger.debug("PsValues.run enter")

        while not self.isInterruptionRequested():
            try:
                self.process_image_received.emit(
                    helper.cm.call_remote_function("ps_values", raise_exception=True)
                )
            except Fault:
                pi.logger.warning("Detected piCtory reset.")
                self.requestInterruption()
                self.driver_reset_detected.emit()
            except Exception as e:
                pi.logger.error(e)
                self.process_image_received.emit(Binary())

            self.msleep(self._cycle_time)

        pi.logger.debug("PsValues.run exit")


class DebugControl(QtWidgets.QWidget, Ui_wid_debugcontrol):
    """Debug controller for main window."""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setupUi(self)

        self.dict_devices = {}
        """Key=position, value=device name."""
        self.dict_ios = {"inp": {}, "out": {}}
        """IO types "inp" "out" which include key=device position, value=list with ios."""
        self.dict_windows = {}
        """Debug IO windows with key=device position, value=DebugIos."""

        self.driver_reset_detected = False
        self.err_workvalues = 0
        self.max_errors = 10
        self.th_worker = PsValues()

        self.vl_devices.addItem(
            QtWidgets.QSpacerItem(20, 1, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        )
        self.cbx_write.setEnabled(False)
        self.cbx_stay_on_top.setChecked(helper.settings.value("debugcontrol/stay_on_top", False, bool))

        self.shc_read_io = QtWidgets.QShortcut(QtGui.QKeySequence("F4"), self)
        self.shc_read_io.setContext(QtCore.Qt.ApplicationShortcut)
        self.shc_read_io.activated.connect(self.on_btn_read_io_pressed)
        self.shc_refresh_io = QtWidgets.QShortcut(QtGui.QKeySequence("F5"), self)
        self.shc_refresh_io.setContext(QtCore.Qt.ApplicationShortcut)
        self.shc_refresh_io.activated.connect(self.on_btn_refresh_io_pressed)
        self.shc_write_o = QtWidgets.QShortcut(QtGui.QKeySequence("F6"), self)
        self.shc_write_o.setContext(QtCore.Qt.ApplicationShortcut)
        self.shc_write_o.activated.connect(self.on_btn_write_o_clicked)

    def __del__(self):
        pi.logger.debug("DebugControl.__del__")

    def _set_gui_control_states(self):
        """Set states depending on acl level."""
        pi.logger.debug("DebugControl._set_gui_control_states")
        # xml_mode view >= 1
        # xml_mode write >= 3
        self.btn_read_io.setEnabled(not self.cbx_write.isChecked())
        self.btn_refresh_io.setEnabled(not self.cbx_refresh.isChecked())
        self.btn_write_o.setEnabled(
            not self.cbx_write.isChecked() and (helper.cm.xml_mode >= 3 or helper.cm.simulating)
        )
        self.cbx_write.setEnabled(
            self.cbx_refresh.isChecked() and (helper.cm.xml_mode >= 3 or helper.cm.simulating)
        )

    def _destroy_io_view(self, device_position=-1):
        """
        Destroy IO view including the button and dict entry.

        :param device_position: Only device position or -1 for all
        """
        pi.logger.debug("DebugControl._destroy_io_view")
        for position in sorted(self.dict_devices) if device_position == -1 else [device_position]:
            if position in self.dict_windows:
                # Remove singe window and button
                win = self.dict_windows[position]  # type: DebugIos
                win.close()
                win.deleteLater()
                win.disconnect()
                del self.dict_windows[position]

            btn = self.gb_devices.findChild((QtWidgets.QPushButton,), str(position))  # type: QtWidgets.QPushButton
            if btn:
                self.vl_devices.removeWidget(btn)
                btn.deleteLater()
                btn.disconnect()

    def _driver_reset_detected(self):
        """Things to do after driver reset."""
        self.driver_reset_detected = True
        self.cbx_write.setChecked(False)
        self.cbx_refresh.setChecked(False)
        for win in self.dict_windows.values():  # type: DebugIos
            win.stat_bar.showMessage(
                self.tr("Driver reset for piControl detected."),
                10000
            )
        self.reload_devices()

    def _work_values(self, refresh=False, write_out=False, process_image=None):
        """
        Read input and output values.

        :param refresh: Refresh unchanged ios from process image
        :param write_out: Write changed outputs to process image
        :param process_image: Use this <class 'Binary'> for work and do not fetch
        """
        if process_image is not None:
            ba_values = process_image
        else:
            try:
                ba_values = helper.cm.call_remote_function("ps_values", raise_exception=True)
            except Fault:
                pi.logger.warning("Detected piCtory reset.")
                self._driver_reset_detected()
                return
            except Exception as e:
                pi.logger.error(e)
                ba_values = Binary()

        # From now on use bytes instead of Binary
        ba_values = bytearray(ba_values.data)

        if not ba_values:
            if self.cbx_refresh.isChecked():
                self.err_workvalues += 1
            else:
                # Raise error on button press
                self.err_workvalues = self.max_errors

            if self.err_workvalues >= self.max_errors:
                for win in self.dict_windows.values():  # type: DebugIos
                    win.stat_bar.setStyleSheet("background-color: red;")
                    win.stat_bar.showMessage(self.tr(
                        "Error while getting values from Revolution Pi."
                    ), 5000)

            return

        if self.err_workvalues > 0:
            self.err_workvalues = 0
            for win in self.dict_windows.values():  # type: DebugIos
                win.stat_bar.setStyleSheet("")

        # Use multicall to set all changed values
        if write_out and helper.cm.connected:
            cli = helper.cm.get_cli()
            xmlmc = MultiCall(cli)
        else:
            xmlmc = []

        for io_type in self.dict_ios:
            for position in self.dict_ios[io_type]:
                if position not in self.dict_windows:
                    continue

                win = self.dict_windows[position]
                for io in self.dict_ios[io_type][position]:  # type: list
                    # ['name', bytelen, byte_address, 'bmk', bitaddress, 'byteorder', signed]
                    # + wordorder since revpipyload 0.9.9
                    value_procimg = bytes(ba_values[io[2]:io[2] + io[1]])
                    if io[4] >= 0:
                        # Bit-IO
                        value_procimg = bool(
                            int.from_bytes(value_procimg, byteorder=io[5], signed=io[6]) & 1 << io[4]
                        )

                    if (refresh or write_out) and io_type == "out":
                        widget_value, last_value = win.get_value(io[0])
                        if widget_value != last_value:
                            # User changed value

                            if not write_out:
                                # Do not write output after change to save this state
                                continue

                            value_procimg = widget_value
                            if type(xmlmc) == MultiCall:
                                xmlmc.ps_setvalue(position, io[0], widget_value)
                            else:
                                # Simulate multicall an collect result to list
                                xmlmc.append(
                                    helper.cm.call_remote_function("ps_setvalue", position, io[0], widget_value)
                                )

                    win.set_value(io[0], value_procimg)

                if self.cbx_refresh.isChecked():
                    win.stat_bar.showMessage(self.tr("Auto update values..."), 1000)
                else:
                    win.stat_bar.showMessage(self.tr("Values updated..."), 2000)

                if self.driver_reset_detected:
                    # Show values, which we can recover to empty process image
                    win.reset_change_value_colors()

        self.driver_reset_detected = False

        # Set values by multi call
        if write_out:
            if isinstance(xmlmc, list):
                self._validate_multicall(xmlmc)
            else:
                self._validate_multicall(xmlmc())

    def _validate_multicall(self, return_list):
        """
        Check xml rpc multi call return values.

        :param return_list: Return values of multi call
        """
        if isinstance(return_list, MultiCallIterator):
            return_list = return_list.results
            if len(return_list) == 0:
                return
        elif not isinstance(return_list, list):
            return
        pi.logger.debug("DebugControl._validate_multicall")

        str_errmsg = ""
        for lst_result in return_list:  # type: list
            # [[device, io, status, msg]] - Yes, double list list :D
            if type(lst_result[0]) == list:
                lst_result = lst_result.pop()
            if not lst_result[2]:
                # Create error message
                device_name = self.dict_devices[lst_result[0]]
                str_errmsg += self.tr(
                    "Error set value of device '{0}' Output '{1}': {2}\n"
                ).format(device_name, lst_result[1], lst_result[3])
            else:
                self.dict_windows[lst_result[0]].reset_change_value_colors(lst_result[1])

        if str_errmsg != "":
            pi.logger.error(str_errmsg)
            if not self.cbx_refresh.isChecked():
                QtWidgets.QMessageBox.critical(self, self.tr("Error"), str_errmsg)

    def deleteLater(self):
        """Clean up all sub windows."""
        pi.logger.debug("DebugControl.deleteLater")

        self.cbx_write.setChecked(False)
        self.cbx_refresh.setChecked(False)
        self._destroy_io_view()

        super().deleteLater()

    def reload_devices(self):
        """Rebuild GUI depending on devices and ios of Revolution Pi."""
        pi.logger.debug("DebugControl.reload_devices")

        if not helper.cm.call_remote_function("psstart", default_value=False):
            # RevPiPyLoad does not support psstart (too old)
            return False

        # ps_devices format: [[0, 'picore01'], [32, 'di01'], ...
        dict_devices = {v[0]: v[1] for v in helper.cm.call_remote_function("ps_devices", default_value=[])}
        if len(dict_devices) == 0:
            # There is no piCtory configuration on the Revolution Pi
            return False

        # Remove not existing or renamed devices
        for position in self.dict_devices:
            if position not in dict_devices or self.dict_devices[position] != dict_devices[position]:
                self._destroy_io_view(position)

        self.dict_devices = dict_devices

        # Format: {position: [['name', bitlength, byte_address, 'bmk', bitaddress, 'byteorder', signed], ...
        inps_data = helper.cm.call_remote_function("ps_inps", default_value=Binary()).data
        outs_data = helper.cm.call_remote_function("ps_outs", default_value=Binary()).data
        if inps_data == b'' or outs_data == b'':
            return False

        dict_inps = pickle.loads(inps_data)
        dict_outs = pickle.loads(outs_data)

        # Take spacer at last position and reinsert it after buttons
        spacer = self.vl_devices.takeAt(self.vl_devices.count() - 1)

        for position in sorted(self.dict_devices):
            if position in self.dict_windows:
                # DebugIos already exists
                if self.dict_windows[position].update_ios(dict_inps[position], dict_outs[position]):
                    # All IOs match the old ones
                    continue
                else:
                    # Destroy old window to build a new one
                    self._destroy_io_view(position)

            win = DebugIos(
                position, self.dict_devices[position],
                dict_inps[position], dict_outs[position]
            )
            win.device_closed.connect(self.on_device_closed)
            self.dict_windows[position] = win

            btn = QtWidgets.QPushButton(self.gb_devices)
            btn.setCheckable(True)
            btn.setObjectName(str(position))
            btn.setText("{0} | {1}".format(position, self.dict_devices[position]))
            btn.clicked.connect(self.on_btn_device_clicked)
            self.vl_devices.addWidget(btn)

        self.vl_devices.addItem(spacer)

        self.dict_ios["inp"] = dict_inps
        self.dict_ios["out"] = dict_outs

        self._work_values(refresh=True)
        self._set_gui_control_states()

        self.cbx_refresh.setChecked(helper.settings.value("debugcontrol/auto_refresh", False, bool))

        return True

    @QtCore.pyqtSlot(bool)
    def on_btn_device_clicked(self, checked: bool):
        """Open or close IO window."""
        pi.logger.debug("DebugControl.on_btn_device_clicked")

        position = int(self.sender().objectName())
        if position in self.dict_windows:
            win = self.dict_windows[position]  # type: QtWidgets.QMainWindow
            win.setWindowFlag(QtCore.Qt.WindowStaysOnTopHint, self.cbx_stay_on_top.isChecked())
            win.setVisible(checked)

    @QtCore.pyqtSlot(int)
    def on_device_closed(self, position: int):
        """Change the check state of button, if window was closed."""
        pi.logger.debug("DebugControl.on_device_closed")
        btn = self.gb_devices.findChild(QtWidgets.QPushButton, str(position))  # type: QtWidgets.QPushButton
        btn.setChecked(False)

    @QtCore.pyqtSlot()
    def on_btn_read_io_pressed(self):
        """Read all IO values and replace changed ones."""
        pi.logger.debug("DebugControl.on_btn_read_io_pressed")
        for win in self.dict_windows.values():  # type: DebugIos
            win.reset_label_colors()
        self._work_values()

    @QtCore.pyqtSlot()
    def on_btn_refresh_io_pressed(self):
        """Read all IO values but do not touch changed ones."""
        pi.logger.debug("DebugControl.on_btn_refresh_io_pressed")
        if not self.cbx_refresh.isChecked():
            self._work_values(refresh=True)

    @QtCore.pyqtSlot()
    def on_btn_write_o_clicked(self):
        """Write outputs."""
        pi.logger.debug("DebugControl.on_btn_write_o_clicked")
        if not self.cbx_write.isChecked() and (helper.cm.xml_mode >= 3 or helper.cm.simulating):
            for win in self.dict_windows.values():  # type: DebugIos
                win.reset_label_colors()
            self._work_values(write_out=True)

    @QtCore.pyqtSlot(int)
    def on_cbx_refresh_stateChanged(self, state: int):
        """Start or stop the auto refresh thread."""
        pi.logger.debug("DebugControl.cbx_refresh_stateChanged")

        # Start / stop worker thread
        if state == QtCore.Qt.Checked and (helper.cm.connected or helper.cm.simulating):
            self.th_worker = PsValues()
            self.th_worker.driver_reset_detected.connect(self._driver_reset_detected)
            self.th_worker.process_image_received.connect(lambda process_image: self._work_values(
                refresh=True,
                write_out=self.cbx_write.isChecked(),
                process_image=process_image
            ))
            self.th_worker.start()

        else:
            self.th_worker.requestInterruption()
            self.th_worker.wait()
            self.th_worker.deleteLater()
            self.cbx_write.setChecked(False)

        self._set_gui_control_states()

    @QtCore.pyqtSlot(bool)
    def on_cbx_refresh_clicked(self, state: bool):
        """Save the state on user action."""
        helper.settings.setValue("debugcontrol/auto_refresh", state)

    @QtCore.pyqtSlot(bool)
    def on_cbx_stay_on_top_clicked(self, state: bool):
        """Save the state on user action."""
        helper.settings.setValue("debugcontrol/stay_on_top", state)

    @QtCore.pyqtSlot(int)
    def on_cbx_write_stateChanged(self, state: int):
        pi.logger.debug("DebugControl.cbx_write_stateChanged")
        checked = state == QtCore.Qt.Checked
        for win in self.dict_windows.values():  # type: DebugIos
            win.write_values = checked

        self._set_gui_control_states()
